#include "CMfield.h"
#include <cmath>
#include <iostream>
#include <string>

using namespace std;
using namespace giac;

map<string,gen> * minpols = 0;

CMfield::CMfield(string s) {

  ten = gen("10",&ct);
  t1 = gen("t1",&ct);
  t2 = gen("t2",&ct);
  
  vars1d.push_back(s__IDNT_e);
  vars1d.push_back(t1);

  vars2d.push_back(s__IDNT_e);
  vars2d.push_back(t1);
  vars2d.push_back(t2);

  nd = 30;
  ndmax = 200;
  ndrecord = 30;
  pf = 10;
  zeronumber = gen("0",&ct);
  onenumber = gen("1",&ct);
  twonumber = gen("2",&ct);

  vecteur e1o, e2o, e3o;
  e1o.push_back(onenumber);
  e1o.push_back(zeronumber);
  e1o.push_back(zeronumber);

  e2o.push_back(zeronumber);
  e2o.push_back(onenumber);
  e2o.push_back(zeronumber);

  e3o.push_back(zeronumber);
  e3o.push_back(zeronumber);
  e3o.push_back(onenumber);

  e1 = gen(e1o);
  e2 = gen(e2o);
  e3 = gen(e3o);

  vecteur zv;
  zv.push_back(zeronumber);
  zv.push_back(zeronumber);
  zv.push_back(zeronumber);
  zerovector = gen(zv);

  vecteur coid(9);
  coid[0] = 1;  coid[1] = 0;    coid[2] = 0;
  coid[3] = 0;  coid[4] = 1;    coid[5] = 0;
  coid[6] = 0;  coid[7] = 0;    coid[8] = 1;
  Id = createMatrix(coid);

  vecteur ro1;
  ro1.push_back(onenumber);
  ro1.push_back(zeronumber);
  vecteur ro2;
  ro2.push_back(zeronumber);
  ro2.push_back(onenumber);
  vecteur ros;
  ros.push_back(ro1);
  ros.push_back(ro2);
  Id2 = gen(ros);

  tag = s;
  if (s.at(0)=='1') {
    tautag = "s4c";
  }
  else {
    if (s.at(2)=='4') {
      tautag = "s4c";
    }
    else {
      if (s.at(2)=='5') {
	tautag = "s5";
      }
      else {
	if (s.at(2)=='1') {
	  tautag = "s1";
	}
	else {
	  cout << "Problem identifying tag..." << endl;
	}
      }
    }
  }
  
  cout << "Sporadic type? " << tautag << endl;

  // Used to encode the CM fields for the Sporadic Groups                                                               
  map<string,gen> m = * data();
  map<string,gen>::iterator it = m.find(s), itend = m.end();
  if (it==itend) {
    cout << "CM field not in data base" << endl;
  }
  else {
    gen te = it->second;
    minpol=te[0];
    cout << "minpol: " << minpol << endl;
    dc = _degree(minpol,&ct);
    
    realminpol=te[3];
    dr = _degree(makesequence(realminpol,s__IDNT_e),&ct);
    
    for (int k=0; is_greater(dr,gen(k),&ct); k++) {
      realminpolcoeffs.push_back(_coeff(makesequence(realminpol,s__IDNT_e,dr-k),&ct));
    }

    if (!2*dr==dc) {
      cout << "This does not look like a CM field, degrees do not match!" << endl;
    }
    
    cout << "real min pol: " << realminpol << endl;
    
    realgen=te[4];
    
    unsigned n = dr.val;
    vars = vecteur(n);
    for (unsigned j=0; j<n; ++j) {
      vars[j] = gen("a"+print_INT_(n-j-1),&ct);
    }
     
    genericreal = _subst(makesequence(_poly2symb(makesequence(vars,s__IDNT_e),&ct),s__IDNT_e,realgen),&ct);
    red(genericreal);

    xv = _evalf(te[1],&ct);
    
    cout << "xv:" << xv << endl;
    cout << "Computing num val of complex/real generators... ";
    computeXval();
      
    cout << "Will now try and identify generator as rootof, this may take some time..." << endl;
    genconj = te[2];
    cout << "Should be conjugate of generator: " << evali(genconj) << endl;
    
    tau = te[5];
    cout << "tau? " << evali(tau) << endl;
    v = te[6];
    cout << "v? " << evali(v) << endl;

    findgenasrootof();
    /*    if (operator_equal(dr,twonumber,&ct)) {
      if (!identifyQuadraticNumber(realgenasrootof,realgenquadnum)) {
	cerr << "Problem identifying real generator as quadratic number..." << endl;
	exit(1);
      }
      else {
	cout << "Real gen as quadratic number: " << realgenquadnum << endl; 
      }
    }
    else {
      vecteur yo;
      gen tee = realgenasrootof;
      vecteur stu;
      int dri = dr.to_int();
      for (int k=1; k<dri; k++) {
	stu.push_back(zeronumber);
      }
      stu.push_back(onenumber);
      cout << "1:   " << stu << endl;
      yo.push_back(stu);
      for (int k=1; k<dri; k++) {
	cout <<  tee << endl;
	gen coe = getCoeffs(tee);
	cout <<  coe << endl;
	vecteur newcoe;
	unsigned le = (*coe._VECTptr).size();
	for (unsigned l=0; l<dri-le; l++) {
	  newcoe.push_back(zeronumber);
	}
	for (unsigned l=0; l<le; l++) {
	  newcoe.push_back(coe[l]);
	}
	cout << "s^" << k << ":   " << newcoe << endl;
	yo.push_back(newcoe);
	tee = _simplify(tee*realgenasrootof,&ct);
      }
      vecteur yo2;
      for (int k=dri-1; k>-1; k--) {
	yo2.push_back(yo[k]);
      }
      yo.clear();
      for (int k=0; k<dri; k++) {
	yo.push_back(_col(makesequence(yo2,k),&ct));
      }
      realgencoeffsinverse = _inverse(gen(yo),&ct);
      cout << "Real gen coeffs inverse matrix: " << realgencoeffsinverse << endl;
      //    }*/
  }
}

void CMfield::increasePrecision() {
  nd = nd+pf;
  if (nd>ndrecord) {
    ndrecord = nd;
  }
  //  cout << "Increasing precision, nd=" << nd << endl;
  prec = _inverse(_pow(makesequence(ten,(int)nd),&ct),&ct);
  computeXval();
}

void CMfield::lowerPrecision() {
  //  cout << "Going back to original precision..." << endl;
  prec = _inverse(_pow(makesequence(ten,(int)nd),&ct),&ct);
  computeXval();
}

bool CMfield::safeSift(gen &gb, vecteur &res) {
  /*
   * I am assuming gb is a RUR Groebner basis, with parameter named s 
   *               (but this does not mean values of realminpol!)
   */
  unsigned savend = nd;
  bool doneSifting = false;
  while (!doneSifting && nd<ndmax) {
    doneSifting = siftAttempt(gb,res);
    if (!doneSifting) {
      increasePrecision();
    }
  }
  nd = savend;
  lowerPrecision();
  return doneSifting;

}

bool CMfield::siftAttempt(gen &gb, vecteur &res) {

  gen sols = _realroot(makesequence(gb[2],prec),&ct);

  res.clear();

  unsigned nbsols = (*sols._VECTptr).size();
  for (unsigned k=0; k<nbsols; k++) {
    gen den = _horner(makesequence(gb[3],sols[k][0],s__IDNT_e),&ct);
    gen svk = _horner(makesequence(gb[4],sols[k][0],s__IDNT_e),&ct)/den;
    /*
     * Beware svi may not be an interval (when rmp is quadratic, say)
     */
    gen J = _intersect(makesequence(svi,svk),&ct);
    if (J.type==3) {
      // To be safe, we will also check that the interval does not contain any other root!
      bool isambiguous = false;
      unsigned l=0;
      unsigned lmax = (*othersvals._VECTptr).size();
      while (!isambiguous && l<lmax) {
	gen Jl = _intersect(makesequence(othersvals[l],svk),&ct);
	isambiguous = (Jl.type==3);
	l++;
      }
      if (isambiguous) {
	return false;
      }
      else {
	res.push_back(sols[k][0]);	
	//	cout <<"RUR parameter " << sols[k][0] << " gives " << endl;
	//	cout << "  s=" << svk << endl;
      }
    }
  }
  return true;
  //  cout << sols << endl;
}

bool CMfield::identifyQuadraticNumber(gen &te, vecteur &res) {
  if (te.is_symb_of_sommet(at_prod)) {
    if (te[1].is_symb_of_sommet(at_plus)) {
      if (te[1][1].is_symb_of_sommet(at_neg)) {
	if (te[1][1][1].is_symb_of_sommet(at_prod)) {
	  if (te[1][1][1][2].is_symb_of_sommet(at_pow)) {
	    res.push_back(te[1][1][1][2][1]);
	    res.push_back(_simplify(-te[1][1][1][1]*te[2],&ct));
	    res.push_back(_simplify(te[1][2]*te[2],&ct));
	    return true;
	  }
	  else {
	    cout << "Don't know what to do (1)..." << endl;
	    return false;
	  }
	}
	else {
	  //		cout << "no *" << endl;
	  if (te[1][1][1].is_symb_of_sommet(at_pow)) {
	    res.push_back(te[1][1][1][1]);
	    res.push_back(_simplify(-te[2],&ct));
	    res.push_back(_simplify(te[1][2]*te[2],&ct));
	    return true;
	  }
	  else {
	    cout << "Don't know what to do (2)..." << endl;
	    return false;
	  }
	}
      }
      else {
	if (te[1][1].is_symb_of_sommet(at_prod)) {
	  if (te[1][1][2].is_symb_of_sommet(at_pow)) {
	    res.push_back(te[1][1][2][1]);
	    res.push_back(_simplify(te[1][1][1]*te[2],&ct));
	    res.push_back(_simplify(te[1][2]*te[2],&ct));
	    return true;
	  }
	}
	else {
	  cout << "te[1][1][0]: " << te[1][1][0] << endl;
	  cout << "Don't know what to do (3)... " << endl;
	  return false;
	}
      }
    }
    else {
      cout << "Don't know what to do (4)..." << endl;
      return false;
    }
  }
  else {
    if (te.is_symb_of_sommet(at_neg)) {
      // could have a minus sign in front, say -a*sqrt(?)
      if (te[1].is_symb_of_sommet(at_prod)) {
	if (te[1][2].is_symb_of_sommet(at_pow)) {
	  res.push_back(te[1][2][1]);
	  res.push_back(-te[1][1]);
	  res.push_back(zeronumber);
	  return true;
	}
	else {
	  return false;
	  cout << "Don't know what to do (5)..." << endl;
	}
      }
      else {
	return false;
	cout << "Don't know what to do (6)..." << endl;
      }
    }
    else {
      // need to double check this one
      if (te.is_symb_of_sommet(at_prod)) {
	if (te[2].is_symb_of_sommet(at_pow)) {
	  res.push_back(te[2][1]);
	  res.push_back(te[1]);
	  res.push_back(zeronumber);
	  return true;
	}
	else {
	  cout << "Don't know what to do (7)..." << endl;
	  return false;
	}
      }
      else {
	if (te.is_symb_of_sommet(at_plus)) {
	  if (te[1].is_symb_of_sommet(at_neg)) {
	    if (te[1][1].is_symb_of_sommet(at_prod)) {
	      if (te[1][1][2].is_symb_of_sommet(at_pow)) {
		res.push_back(te[1][1][2][1]);
		res.push_back(-te[1][1][1]);
		res.push_back(te[2]);
		return true;
	      }
	      else {
		cout << "Don't know what to do (8)..." << endl;
		return false;
	      }
	    }
	    else {
	      if (te[1][1].is_symb_of_sommet(at_pow)) {
		res.push_back(te[1][1][1]);
		res.push_back(-1);
		res.push_back(te[2]);
		return true;
	      }
	      else {
		cout << "Don't know what to do (9)..." << endl;
		return false;
	      }
	    }
	  }
	}
      }
    }
  }
}

gen CMfield::getCoeffs(gen &te) {
  if (te.is_symb_of_sommet(at_prod)) {
    return _simplify(te[1][1]*te[2],&ct);
  }
  else {
    return te[1];
  }
}

bool CMfield::convertRootOf(gen &te, gen &res) {
  /*
   * The goal is to convert an expression involving rootofs of the form rootof(p,realminpol)
   *   into polynomial expressions in s... if it works, puts this polynomial in res
   */
  if (operator_equal(te,zeronumber,&ct)) {
    res = zeronumber;
    return true;
  }
  else {
    if (te.is_symb_of_sommet(at_prod)) {
      /*      cout << "Produit? " << endl;
      cout << te[0] << endl;
      cout << te[1] << endl;
      cout << te[2] << endl;*/
      if (te[1].is_symb_of_sommet(at_rootof)) {
	/*	cout << te[1][0] << endl;
	cout << te[1][1] << endl;
	cout << te[1][2] << endl;*/
	if (!operator_equal(te[1][2],realminpolcoeffs,&ct)) {
	  cout << "Rootof base poly is not realminpol, this value may not be in base field..." << endl;
	  //	  cout << "Realminpolsymb: " << realminpolcoeffs[1] << endl;
	  return false;
	}
	else {	
	  //	  gen co = _simplify(te[1][1]*te[2],&ct);
	  //	  cout << "Coeffs are: " << co << endl;
	  //  want    te = ak*s^k+...+a1*s+a0 (k=dr-1?)
	  // then need to solve a linear system
	  gen yoyo = _simplify(realgencoeffsinverse*_simplify(te[1][1]*te[2],&ct),&ct);
	  res = _horner(makesequence(yoyo,s__IDNT_e),&ct);
	  //	  cout << "value: " << res << endl;
	  //	  cout << "Numerical: " << evali(res) << endl;
	  return true;
	}
      }
      else {
	// may still be OK if quadratic extension...
	if (operator_equal(dr,twonumber,&ct)){
	  //	  cout << "Quadratic, trying to analyze " << te[1]<< endl;
	  vecteur qn;
	  if (identifyQuadraticNumber(te,qn)) {
	    //	    cout << "Result: " << qn << endl;
	    gen c = _simplify(qn[1]/realgenquadnum[1],&ct);
	    gen d = _simplify(qn[2]-realgenquadnum[2]*c,&ct);
	    res = (c*s__IDNT_e+d); 
	    cout << res << endl;
	    //	    cout << "Check: " << evali(res) << " VS " << _evalf(te,&ct) << endl;
	    return true;
	  }
	  else {
	    return false;
	  }
	}
	else {
	  return false;
	  //	  cout << "Don't know how to convert, this does not look like an expression I can analyze automatically..." << endl;
	}
      }
    }
    else {
      if (te.is_symb_of_sommet(at_rootof)) {
	/*	cout << "Rootof? " << endl;
	cout << te[0] << endl;
	cout << te[1] << endl;
	cout << te[2] << endl; */   
	if (!operator_equal(te[2],realminpolcoeffs,&ct)) {
	  cout << "Rootof base poly is not realminpol, this value may not be in base field..." << endl;
	  return false;
	}
	else {
	  //	  cout << "Coeffs are: " << te[1] << endl;
	  //  want    te = ak*s^k+...+a1*s+a0 (k=dr-1?)
	  // then need to solve a linear system
	  gen yoyo = _simplify(realgencoeffsinverse*te[1],&ct);
	  res = _horner(makesequence(yoyo,s__IDNT_e),&ct);
	  //	  cout << "value: " << res << endl;
	  //	  cout << "Numerical: " << evali(res) << endl;
	}
      }
      else {
	if (operator_equal(dr,twonumber,&ct)){
	  //	  cout << "Quadratic, trying to analyze " << te << endl;
	  vecteur qn;
	  if (identifyQuadraticNumber(te,qn)) {
	    //	    cout << "Result: " << qn << endl; 
	    gen c = _simplify(qn[1]/realgenquadnum[1],&ct);
	    gen d = _simplify(qn[2]-realgenquadnum[2]*c,&ct);
	    res = (c*s__IDNT_e+d); 
	    cout << res << endl;
	    //	    cout << "Check: " << evali(res) << " VS " << _evalf(te,&ct) << endl;
	    return true;
	  }
	  else {
	    //	    cout << "Don't know what to do..." << endl;
	    return false;
	  }
	}
	else {
	  //	  cout << "Don't know how to convert, this does not look like an expression I can analyze automatically..." << endl;
	  return false;
	}
      }
    }
  }
}

gen CMfield::findInField(gen g, gen approx) {

  vecteur variables = lidnt(g);
  //  cout << "Variables: " << variables << endl;
  if (variables.size()!=1) {
    cout << "Polynomial is not univariate?" << endl;
  }
  else {

    gen fa = _factors(makesequence(g,genasrootof),&ct);
  
    //    cout << "fa=" << fa << endl;

    int nf = (*fa._VECTptr).size()/2;
    //    cout << "Nb factors: " << nf << endl;  

    int k=0;
    bool found = false;
    while (k<nf && !found) {
      gen df = _degree(makesequence(fa[2*k],variables[0]),&ct);
      if (df==1) {
	//	cout << "Factor: " << fa[2*k] << endl;
	gen te = _simplify(-_coeff(makesequence(fa[2*k],variables[0],0),&ct)/_coeff(makesequence(fa[2*k],variables[0],1),&ct),&ct);
	//	cout << _evalf(te,&ct) << endl;
	found = is_greater(1e-6,abs(_evalf(te,&ct)-approx),&ct);
        if (found) {
	  if (te.is_symb_of_sommet(at_prod)) {
	    gen subfa = te._SYMBptr->feuille;
	    gen yo = _simplify(te[1][1]*te[2],&ct);
	    int umax = dc.val-(*yo._VECTptr).size();
	    for (int u=0; u<umax; u++) {
	      (*yo._VECTptr).insert((*yo._VECTptr).begin(),zeronumber);
	    }
	    gen ls = _inverse(genasrootofpowers,&ct)*yo;
	    return _poly2symb(makesequence(ls,x__IDNT_e),&ct);
	    // return _poly2symb(makesequence(_linsolve(makesequence(genasrootofpowers,yo),&ct),x__IDNT_e),&ct);
	  }
	  else {
	    if (te.is_symb_of_sommet(at_rootof)) {
	      gen yo = _simplify(te[1],&ct);
	      //	      gen res = _poly2symb(makesequence(yo,x__IDNT_e),&ct);
	      int umax = dc.val-(*yo._VECTptr).size();
	      for (int u=0; u<umax; u++) {
		(*yo._VECTptr).insert((*yo._VECTptr).begin(),zeronumber);
	      }
	      gen ls = _inverse(genasrootofpowers,&ct)*yo;
	      return _poly2symb(makesequence(ls,x__IDNT_e),&ct);
	      //	      return _poly2symb(makesequence(_linsolve(makesequence(genasrootofpowers,yo),&ct),x__IDNT_e),&ct);
	    }
	    else {
	      /*	      vecteur res;
	      if (identifyQuadraticNumber(g,res)) {
		cout << res << endl;
	      }
	      else {
		cout << "Could not identify quadratic number..." << endl;
		}*/
	      cout << "Problem reading factors!" << endl;
	    }
	  }
	}
      }
      k++;
    }
    if (!found) {
      cout << "Could not find this solution in number field..." << endl;
    }
  }
}

void CMfield::findgenasrootof() {
  gen xtest("xx",&ct);
  gen rtx=rootof(minpol,&ct);
  sto(xtest,rtx,&ct);
  gen fa = _factors(makesequence(minpol,rtx),&ct);
  int nf = (*fa._VECTptr).size()/2;
  int k=0;
  bool found = false;
  while (k<nf && !found) {
    gen df = _degree(makesequence(fa[2*k],x__IDNT_e),&ct);
    if (df==1) {
      gen te = _simplify(-_coeff(makesequence(fa[2*k],x__IDNT_e,0),&ct)/_coeff(makesequence(fa[2*k],x__IDNT_e,1),&ct),&ct);
      found = 1e-8 > abs(_evalf(te,&ct)-xv);
      if (found) {
	genasrootof = te;
	cout << "Gen as rootof: " << genasrootof << endl;
	cout << " check: " << _evalf(genasrootof,&ct) << endl;
	cout << "        " << _evalf(xvi,&ct) << endl;
	int m = dc.val;
	vecteur gpv;
	vecteur col0;
	for (unsigned k=0; k<m-1; k++) {
	  col0.push_back(zeronumber);
	}
	col0.push_back(onenumber);
	for (unsigned k=0; k<m-1; k++) {
	  if (te.is_symb_of_sommet(at_prod)) {
	    gen subfa = te._SYMBptr->feuille;
	    gen yo = _simplify(te[1][1]*te[2],&ct);
	    int umax = m-(*yo._VECTptr).size();
	    for (int u=0; u<umax; u++) {
	      (*yo._VECTptr).insert((*yo._VECTptr).begin(),zeronumber);
	    }
	    gpv.insert(gpv.begin(),yo);
	  }
	  else {
	    if (te.is_symb_of_sommet(at_rootof)) {
	      gen yo = _simplify(te[1],&ct);
	      int umax = m-(*yo._VECTptr).size();
	      for (int u=0; u<umax; u++) {
		(*yo._VECTptr).insert((*yo._VECTptr).begin(),zeronumber);
	      }
	      gpv.insert(gpv.begin(),yo);
	    }
	    else {
	      cout << "Problem with rootof powers!" << endl;
	    }
	  }	  
	  //	  cout << te << endl;
	  //	  cout << gpv[gpv.size()-1] << endl;
	  te = _simplify(te*genasrootof,&ct);
	}
	gpv.push_back(col0);
	genasrootofpowers = _tran(gen(gpv),&ct);
	//	cout << "Gen powers: " << endl;
	//	cout << genasrootofpowers << endl;
      }
    }
    k++;
  }
  if (!found) {
    cout << "Problem computing field generator as rootof!" << endl;
  }  
  else {
    gen stest("ss",&ct);
    gen rt=rootof(realminpol,&ct);
    sto(stest,rt,&ct);

    fa = _factors(makesequence(realminpol,rt),&ct);
    //    fa = _factors(makesequence(realminpol,rootof(realminpol,&ct)),&ct);
    //    cout << "fa: " << fa << endl;
    nf = (*fa._VECTptr).size()/2;
    k=0;
    found = false;
    while (k<nf && !found) {
      gen df = _degree(makesequence(fa[2*k],s__IDNT_e),&ct);
      if (operator_equal(df,onenumber,&ct)) {
	gen te = _simplify(-_coeff(makesequence(fa[2*k],s__IDNT_e,0),&ct)/_coeff(makesequence(fa[2*k],s__IDNT_e,1),&ct),&ct);
	//		cout << "factor: " << fa[2*k] << endl;
	//		cout << "te=" << te << endl;
	//		cout << " num val: " << abs(_evalf(te,&ct)-svi) << endl;
	found = is_greater(1e-8,abs(_evalf(te,&ct)-svi),&ct);
	//	found = 1e-8 > abs(_evalf(te,&ct)-svi);
	if (found) {
	  realgenasrootof = te;
	  cout << "Real gen as rootof: " << realgenasrootof << endl;
	  cout << " check: " << _evalf(realgenasrootof,&ct) << endl;
	  cout << "        " << _evalf(svi,&ct) << endl;
	}
      }
      k++;
    }
    if (!found) {
      cout << "Problem computing real subfield generator as rootof!" << endl;
    }  
  }
}

gen CMfield::convertToReal(gen g) {  
  gen eqs = _symb2poly(makesequence(genericreal-g,x__IDNT_e),&ct);
  gen sol = _solve(makesequence(eqs,vars),&ct);
  //  cout << "sol: " << sol << endl;
  if ((*sol._VECTptr).size()==1) {
    return  _simplify(_poly2symb(makesequence(sol[0],s__IDNT_e),&ct),&ct);
  }
  else {
    cout << "Error converting to real generator, is this number real??" << endl; 
  }
}

void CMfield::computeXval() {

  gen rtx = _complexroot(makesequence(minpol,prec),&ct);

  int counter = 0;
  int k = 0;
  int kmax = (*rtx._VECTptr).size();
  while (k<kmax) {
    gen a = abs(rtx[k][0] - xv);
    if (1e-10>a) {
      counter++;
      xvi = rtx[k][0];
    }
    k++;
  }
  /*  if (counter!=1) {
    cout << "Found " << counter << " values for x, this should not happen!" << endl;
  }
  else {
    cout << "Generator for field: " << _eval(xvi,&ct) << endl;
    }*/

  gen val = _evalf(_horner(makesequence(realgen,xv,x__IDNT_e),&ct),&ct);

  gen vali = val.im(&ct);
  if (_abs(vali,&ct)>1e-10) {
    cout << "Please check real generator, it does not seem real.." << endl;
  }
  else {
    gen valr = val.re(&ct);
    
    gen rts = _realroot(makesequence(realminpol,prec),&ct);
    vecteur osv;
    
    int counter = 0;
    int k = 0;
    int kmax = (*rts._VECTptr).size();

    while (k<kmax) {
      //      cout << rts[k] << endl;
      //      cout << valr << endl;
      gen a = abs(rts[k][0] - valr);
      //      cout << "a=" << a << endl;
      if (is_greater(1e-8,a,&ct)) {
	counter++;
	if (rts[k][0].type==8) {
	  svi = convert_interval(rts[k][0],nd,&ct);	  	  
	}
	else {
	  svi = rts[k][0];
	  //	  osv.push_back(rts[k][0]);
	}	
      }
      else {
	/*	if (operator_equal(_left(rts[k][0],&ct),rts[k][0],&ct)) {
	  osv.push_back(convert_interval(rts[k][0],nd,&ct));	  	  
	}
	else {
	  osv.push_back(rts[k][0]);
	  }*/	

	if (rts[k][0].type==8) {
	  osv.push_back(convert_interval(rts[k][0],nd,&ct));	  
	}
	else {
	  osv.push_back(rts[k][0]);
	}

      }
      k++;
    }
    if (counter!=1) {
      cout << "Found " << counter << " values for s, this should not happen!" << endl;
    }
    else {
      //      cout << "Generator for real subfield: " << svi;
      /*     if (operator_equal(_left(svi,&ct),svi,&ct)) {
	svi = convert_interval(svi,nd,&ct);
	}*/
      if (svi.type==8) {
	svi = convert_interval(svi,nd,&ct);
      }

      othersvals = gen(osv);
      //     cout << "Other real roots of real min pol: " << othersvals << endl;
    }
  }

}

gen CMfield::createMatrix(unsigned m, unsigned n, vecteur v) {
  vecteur te;
  for (unsigned j=0; j<m; j++) {
    vecteur tej;
    for (unsigned k=0; k<n; k++) {
      tej.push_back(v[m*j+k]);
    }
    te.push_back(tej);
  }
  return gen(te);
}

gen CMfield::createMatrix(vecteur v) {
  vecteur r1 = vecteur(3);
  vecteur r2 = vecteur(3);
  vecteur r3 = vecteur(3);
  for (int j=0; j<3; j++) {
    r1[j] = v[j];  
  }
  for (int j=0; j<3; j++) {
    r2[j] = v[3+j];  
  }
  for (int j=0; j<3; j++) {
    r3[j] = v[6+j];  
  }
  vecteur res(3);
  res[0] = gen(r1);
  res[1] = gen(r2);
  res[2] = gen(r3);
  return gen(res);
}

gen CMfield::mypower(gen M, unsigned k) {
  gen res = Id;
  for (int j=0; j<k; j++) {
    res = res*M;
    matred(res);
  }
  return res;
}

int CMfield::order(gen M, unsigned k) {
  bool done = isScalar(M);
  gen te = M;
  unsigned j=1;
  if (done) {
    return j;
  }
  while (!done && j<k) {
    j++;
    te = te * M;
    matred(te);
    done = isScalar(te);
  }
  if (done) {
    return j;
  }
  else {
    return -1;
  }
}

int CMfield::order(gen M) {
  int k = order(M,1000);
  if (k>-1) {
    return k;
  }
  else {
    //    cout << "Order is > 1000" << endl;
    return -1;
  }
}

int CMfield::linearorder(gen M, unsigned k) {
  bool done = operator_equal(M,Id,&ct);
  gen te = M;
  unsigned j=1;
  if (done) {
    return j;
  }
  while (!done && j<k) {
    j++;
    te = te * M;
    matred(te);
    done = operator_equal(M,Id,&ct);
  }
  if (done) {
    return j;
  }
  else {
    return -1;
  }
}

int CMfield::linearorder(gen M) {
  int k = order(M,1000);
  if (k>-1) {
    return k;
  }
  else {
    //    cout << "Order is > 1000" << endl;
    return -1;
  }
}

int CMfield::braidOrder(gen M, gen N) {
  int k = braidOrder(M,N,1000);
  if (k>-1) {
    return k;
  }
  else {
    //    cout << "Order is > 1000" << endl;
    return -1;
  }  
}

bool CMfield::areDep(gen v1, gen v2) {
  bool allzero = true;
  unsigned c = 0;
  while (allzero && c<3) {
    gen te = v1[c]*v2[(c+1)%3] - v2[c]*v1[(c+1)%3]; 
    red(te);
    allzero = operator_equal(te,zeronumber,&ct);
    c++;
  }
  return allzero;
}

bool CMfield::isMultiple(gen M, gen N) {
  bool res = true;
  unsigned c = 0;
  while (res && c<3) {
    res = areDep(M[c],N[c]);
    c++;
  }
  return res;
}

int CMfield::braidOrder(gen M, gen N, unsigned k) {
  gen A = M;
  gen B = N;
  if (isMultiple(A,B)) {
    return 1;
  }
  else {
    int o = 2;
    while (o<k)  {
      A = A*N;
      B = B*M;
      matred(A);
      matred(B);
      if (isMultiple(A,B)) {
	return o;
      }
      else {
	o++;
	A = A*M;
	B = B*N;
	matred(A);
	matred(B);
	if (isMultiple(A,B)) {
	  return o;
	}
	else {
	  o++;
	}
      }
    }
  }
  return -1;
}

bool CMfield::testDependent(gen V, gen W) {
  bool res = false;
  unsigned k=0;
  while (!res && k<3) {
    unsigned kp1 = (k+1)%3;
    gen te = V[k]*W[kp1] - V[kp1]*W[k];
    red(te);
    res = operator_equal(te,zeronumber,&ct);
  }
}

bool CMfield::isScalar(gen M) {
  if (M[0][0]==M[1][1] && M[1][1]==M[2][2]) {
    gen ze = gen("0",&ct);
    return (M[0][1]==ze && M[0][2]==ze && M[1][2]==ze && M[1][0]==ze && M[2][0]==ze && M[2][1]==ze);
  }
  else {
    return false;
  }
}

void CMfield::red(gen &g) {
  g = _simplify(g,&ct);
  if (contains(g,s__IDNT_e)) {
      if (contains(g,x__IDNT_e)) {
	gen gx = _simplify(_subst(makesequence(g,s__IDNT_e,realgen),&ct),&ct);
	g = _simplify(_rem(makesequence(gx,minpol,x__IDNT_e),&ct),&ct);
      }
      else {
	g = _simplify(_rem(makesequence(g,realminpol,s__IDNT_e),&ct),&ct);
      }
  }
  else {
    g= _simplify(_rem(makesequence(g,minpol,x__IDNT_e),&ct),&ct);
  }
}

void CMfield::vectred(gen &v) {
  unsigned n = (*v._VECTptr).size();
  for (int j=0; j<n; j++) {
    red((*v._VECTptr)[j]);
  }
}

void CMfield::matred(gen &M) {
  unsigned m = (*M._VECTptr).size();
  for (int j=0; j<m; j++) {
    vectred((*M._VECTptr)[j]);
  }
}

gen CMfield::myconj(gen g) {
  gen gco = _subst(makesequence(g,x__IDNT_e,genconj),&ct);
  red(gco);
  return gco;
}

gen CMfield::vectconj(gen v) {
  unsigned n = (*v._VECTptr).size();
  vecteur res(n);
  for (int j=0; j<n; j++) {
    res[j] = myconj(v[j]);
  }
  return gen(res);
}

gen CMfield::matconj(gen M) {
  unsigned m = (*M._VECTptr).size();
  vecteur res(m);
  for (int j=0; j<m; j++) {
    res[j] = vectconj(M[j]);
  }
  return gen(res);
}

gen CMfield::det(gen M) {
  gen te = _det(M,&ct);
  red(te);
  return te;
}

gen CMfield::trace(gen M) {
  gen te = M[0][0]+M[1][1]+M[2][2];
  red(te);
  return te;
}

gen CMfield::discr(gen z) {
  gen zc = myconj(z);
  gen z2 = z*z;
  gen z2c = myconj(z2);
  gen z3 = z2*z;
  gen z3c = myconj(z3);
  gen te = z2*z2c - 4*(z3+z3c) + 18*z*zc - 27;
  red(te);
  return convertToReal(te);
}

gen CMfield::discrAdjusted(gen M) {

  /*  gen nu = det(M);
  red(nu);
  cout << "Det M by xcas: " << nu << endl;
  cout << evali(nu) << endl;*/

  /*  gen tee = zeronumber;
  tee = tee + M[0][0]*M[1][1]*M[2][2];
  red(tee);
  tee = tee + M[0][1]*M[1][2]*M[2][0];
  red(tee);
  tee = tee + M[1][0]*M[2][1]*M[0][2];
  red(tee);
  tee = tee - M[2][0]*M[1][1]*M[0][2];
  red(tee);
  tee = tee - M[1][0]*M[0][1]*M[2][2];
  red(tee);
  tee = tee - M[2][1]*M[1][2]*M[0][0];
  red(tee);
  cout << "COMPUTING DET BY HAND... " << tee << endl;*/
  gen tee = det(M);
  red(tee);

  gen de = inverse(tee);

  gen z = trace(M);

  gen zc = myconj(z);

  gen z2 = z*z;
  gen z2c = myconj(z2);

  gen z3 = z2*z*de;
  gen z3c = myconj(z3);

  gen te = z2*z2c - 4*(z3+z3c) + 18*z*zc - 27;
  red(te);
  return convertToReal(te);
}

bool CMfield::isReal(gen g) {
  gen gt = g-myconj(g);
  red(gt);
  return gt==0;
}

gen CMfield::realPart(gen g) {
  gen gt = (g+myconj(g))/2;
  red(gt);
  return convertToReal(gt);
}

gen CMfield::inverse(gen g) {
  if (contains(g,s__IDNT_e)) {
    if (contains(g,x__IDNT_e)) {
      gen gx = _subst(makesequence(g,s__IDNT_e,realgen),&ct);
      gen e =_egcd(makesequence(gx,minpol,x__IDNT_e),&ct);
      gen res = e[0]/e[2];
      red(res);
      return res;
    }
    else {
      gen e =_egcd(makesequence(g,realminpol,s__IDNT_e),&ct);
      gen res = e[0]/e[2];
      red(res);
      return res;
    }
  }
  else {
    if (contains(g,x__IDNT_e)) {
      gen e =_egcd(makesequence(g,minpol,x__IDNT_e),&ct);
      gen res = e[0]/e[2];
    //    cout << "e[2]=" << e[2] << endl;
    //    cout << "     " << evali(e[2]) << endl;
      red(res);
      return res;
    }
    else {
      return onenumber/g; // should test if nonzero?
    }
  }  
}

gen CMfield::evali(gen g) {
  if (contains(g,s__IDNT_e)) {
    if (contains(g,x__IDNT_e)) {
      gen gx = _subst(makesequence(g,s__IDNT_e,realgen),&ct);
      return _subst(makesequence(gx,x__IDNT_e,xvi),&ct);
    }
    else {
      return _subst(makesequence(g,s__IDNT_e,svi),&ct);
    }
  }
  else {
    if (contains(g,x__IDNT_e)) {
      return _subst(makesequence(g,x__IDNT_e,xvi),&ct);
    }
    else {
      // this will work only for numbers?!
      if (operator_equal(g,zeronumber,&ct)) {
	return convert_interval(onenumber,nd,&ct)-onenumber;
      }
      else {
	return convert_interval(g,nd,&ct);
      }
    }
  }  
}

gen CMfield::vectevali(gen v) {
  vecteur te;
  unsigned n = (*v._VECTptr).size();
  for (int j=0; j<n; j++) {
    te.push_back(evali((*v._VECTptr)[j]));
  }
  gen tee = gen(te);
  return tee;
}

gen CMfield::matevali(gen M) {
  vecteur te;
  unsigned m = (*M._VECTptr).size();
  for (int j=0; j<m; j++) {
    te.push_back(vectevali((*M._VECTptr)[j]));
  }
  gen tee = gen(te);
  return tee;
}

gen CMfield::evaln(gen g) {
  if (contains(g,s__IDNT_e)) {
    if (contains(g,x__IDNT_e)) {
      gen gx = _subst(makesequence(g,s__IDNT_e,realgen),&ct);
      return _subst(makesequence(gx,x__IDNT_e,xv),&ct);
    }
    else {
      return _subst(makesequence(g,s__IDNT_e,sv),&ct);
    }
  }
  else {
    if (contains(g,x__IDNT_e)) {
      return _subst(makesequence(g,x__IDNT_e,xv),&ct);
    }
    else {
      // this will work only for numbers?!
      if (operator_equal(g,zeronumber,&ct)) {
	return g;
      }
      else {
	return _evalf(makesequence(g,(int)nd),&ct);
      }
    }
  }  
}

gen CMfield::vectevaln(gen v) {
  vecteur te;
  unsigned n = (*v._VECTptr).size();
  for (int j=0; j<n; j++) {
    te.push_back(evaln((*v._VECTptr)[j]));
  }
  gen tee = gen(te);
  return tee;
}

gen CMfield::matevaln(gen M) {
  vecteur te;
  unsigned m = (*M._VECTptr).size();
  for (int j=0; j<m; j++) {
    te.push_back(vectevaln((*M._VECTptr)[j]));
  }
  gen tee = gen(te);
  return tee;
}

bool CMfield::checkSign(gen g, int &res){
  unsigned savend = nd;
  bool done = false;
  while (!done && nd<ndmax) {
    done = checkSignAttempt(g,res);
    if (!done) {
      increasePrecision();
    }
  }
  nd = savend;
  lowerPrecision();
  return done;
}

gen CMfield::transpose(gen M) {
  vecteur coo;
  for (unsigned j=0; j<3; j++) {
    for (unsigned k=0; k<3; k++) {
      coo.push_back(M[k][j]);
    }
  }
  return createMatrix(coo);
}

bool CMfield::checkSignAttempt(gen g, int &sign) {
  // We will assume g is a polynomial in s, and it is reduced...
  //   (this may not be a good idea...)
  if (operator_equal(g,zeronumber,&ct)) {
    sign = 0;
    return true;
  }
  else {
    gen va = _subst(makesequence(g,s__IDNT_e,svi),&ct);
    //    if (operator_equal(_left(va,&ct),va,&ct)) {
    if (va.type==8) {
      //      cout << "SHOULD CHECK THIS" << endl;
      //      cout << "va=" << va << endl;
      //      if (is_greater(zeronumber,va,&ct)) {
      if (is_greater(-prec,va,&ct)) {
	sign = -1;
	return true;
      }
      else {
	if (is_greater(va,prec,&ct)) {
	  sign = 1;
	  return true;
	}
	else {
	  if (va==zeronumber)  {
	    sign = 0;
	    return true;
	  }
	  else {
	    return false;
	  }
	}
      }
    }
    else {
      //      cout << "(interval) va=" << va << endl;
      if (is_greater(zeronumber,_right(va,&ct),&ct)) {
	sign = -1;
	return true;
      }
      else {
	if (is_greater(_left(va,&ct),zeronumber,&ct)) {
	  sign = 1;
	  return true;
	}
	else {
	  return false;
	}
      }         
    }
  }
}

bool CMfield::checkSignPol(gen par, gen mpol, gen q, int &res) {

  if (operator_equal(zeronumber,q,&ct)) {
    res = 0;
    return true;
  }
  
  unsigned savend = nd;
  bool worked = false;
  while (!worked && nd<ndmax) {
    worked = checkSignPolAttempt(par,mpol,q,res);
    if (!worked) {
      increasePrecision();
    }
  }
  nd = savend;
  lowerPrecision();
  return worked;
}

bool CMfield::checkSignPolAttempt(gen par, gen mpol, gen q, int &res) {
  // u is an expression in s = gen of totally real subfield                                                                        //   par is an interval, containing a single root of mpol                                                                        //   we want to evaluate the sign of q at that root                                                                            

  //  cout << "Precision: " << prec << endl;

  gen sols = _realroot(makesequence(mpol,prec),&ct);
  unsigned nbsols = (*sols._VECTptr).size();
  int count = 0;
  int ksave;
  for (unsigned k=0; k<nbsols; k++) {
    // beware sols[k][0] may not be an interval...
    gen J = _intersect(makesequence(par,sols[k][0]),&ct);
    //   cout << "J=" << J << endl;
    if (J.type==3) {
      count++;
      ksave = k;
      cout << "ksave = " << ksave << endl;
    }
  }
  if (count!=1) {
    cerr << "Interval is not isolating, please double check!" << endl;
    exit(1);
  }
  else {

    //    gen te = _horner(makesequence(q,convert_interval(par_refined,nd,&ct),s__IDNT_e),&ct);
    cout << "nd: " << nd << endl;
    cout << "prec: " << prec << endl;
    cout << "mpol: " << mpol << endl;
    cout << "New interval: " << sols[ksave][0] << endl;
    gen te = _horner(makesequence(q,sols[ksave][0],s__IDNT_e),&ct);
    //    gen te = _subst(makesequence(q,s__IDNT_e,sols[ksave][0]),&ct);
    cout << "te=" << te << endl;

    if (is_greater(zeronumber,_right(te,&ct),&ct)) {
      res = -1;
      return true;
    }
    else {
      if (is_greater(_left(te,&ct),zeronumber,&ct)) {
        res = 1;
        return true;
      }
      else {
        // need to check if it is zero!                                                                                         
	gen e = _egcd(makesequence(mpol,q,s__IDNT_e),&ct);
	//	cout << "e=" << e << endl;
        gen cr = _realroot(makesequence(e[2],prec),&ct);
        for (unsigned k=0; k<(*cr._VECTptr).size(); k++) {
	  //	  cout << "cr[" << k << "][0]=" << cr[k][0] << endl;
          gen J = _intersect(makesequence(sols[ksave][0],cr[k][0]),&ct);
          if (J.type==3) {
	    res = 0;
            return true;
          }
        }
        // if we get here, then the expression is non-zero but we don't know its sign...                                      
        return false;
      }
    }
  }

}

gen CMfield::matinverse(gen M) {
  gen de = det(M);
  red(de);
  if (operator_equal(de,zeronumber,&ct)) {
    cerr << "Matrix is not invertible" << endl;
    exit(1);
  }
  else {
    vecteur co;
    co.push_back(M[1][1]*M[2][2] - M[1][2]*M[2][1]);
    co.push_back(- M[0][1]*M[2][2] + M[0][2]*M[2][1]);
    co.push_back(M[0][1]*M[1][2] - M[0][2]*M[1][1]);

    co.push_back(- M[1][0]*M[2][2] + M[1][2]*M[2][0]);
    co.push_back(M[0][0]*M[2][2] - M[0][2]*M[2][0]);
    co.push_back(- M[0][0]*M[1][2] + M[0][2]*M[1][0]);

    co.push_back(M[1][0]*M[2][1] - M[1][1]*M[2][0]);
    co.push_back(- M[0][0]*M[2][1] + M[0][1]*M[2][0]);
    co.push_back(M[0][0]*M[1][1] - M[0][1]*M[1][0]);
    //    cout << "co: " << co << endl;

    gen A = createMatrix(co);
    A = A*inverse(de);
    matred(A);
    return A;
  }
}


map<string,gen> * CMfield::data() {
  /*
   * #0: minimal polynomial of generator
   * #1: numerical value of that gen
   * #2: expr of complex conjugate of gen in terms of gen
   * #3: min pol of gen of maximal totally real subfield
   * #4: expression of real generator in terms of complex generator
   * #5: expr of tau in terms of complex generator
   * #6: expr of v in terms of complex gen, a root of unity
   */
  if (!minpols) {
    minpols = new map<string,gen>;
    (*minpols)["3s4c"] = gen(makesequence( gen("x^4+5*x^2+1",&ct),
					   gen("i*(sqrt(3)+sqrt(7))/2",&ct),
					   gen("-x",&ct),
					   gen("s^2+10*s+4",&ct), 
					   gen("2*x^2",&ct),
					   gen("(-x^3-6*x-1)/2",&ct),
					   gen("(-x^3-4*x-1)/2",&ct) ));

    (*minpols)["4s4c"] = gen(makesequence( gen("x^4-2*x^3+7*x^2-6*x+2",&ct), 
					   gen("1/2+i*(1+sqrt(7)/2)",&ct),
					   gen("-x+1",&ct),
					   gen("s^2+10*s-3",&ct), 
					   gen("2*x^2-2*x+1",&ct),
					   gen("(-2*x^3+3*x^2-14*x+5)/3",&ct),
					   gen("(-2*x^3+3*x^2-11*x+5)/3",&ct) ));

    (*minpols)["5s4c"] = gen(makesequence( gen("x^8-2*x^7+10*x^6-12*x^5+ 29*x^4-28*x^3+15*x^2-13*x+11",&ct), 
					   gen(".809016994374945+2.27393217182745*i",&ct),
					   gen("-258/3091*x^7+466/3091*x^6-2298/3091*x^5+2435/3091*x^4-5836/3091*x^3+5470/3091*x^2-2762/3091*x+241/281",&ct),
					   gen("s^4+16*s^3+61*s^2-24*s-59",&ct), 
					   gen("-158/3091*x^7-98/3091*x^6-976/3091*x^5-857/3091*x^4-2328/3091*x^3+3254/3091*x^2+1735/3091*x+6/281",&ct),
					   gen("(-5040*x^7+8025*x^6-48701*x^5+42895*x^4-145275*x^3+94995*x^2-112900*x+31372)/58729",&ct),
					   gen("(-5040*x^7+8025*x^6-48701*x^5+42895*x^4-145275*x^3+94995*x^2-54171*x+31372)/58729",&ct) ));

    (*minpols)["6s4c"] = gen(makesequence( gen("x^4-4*x^3+11*x^2-14*x+7",&ct),
					   gen("1+i*(sqrt(3)+sqrt(7))/2",&ct),
					   gen("-x+2",&ct),
					   gen("s^2+6*s-12",&ct), 
					   gen("2*x^2-4*x+4",&ct),
					   gen("(-x^3+3*x^2-9*x+6)/2",&ct),
					   gen("(-x^3+3*x^2-7*x+6)/2",&ct) ));

    (*minpols)["8s4c"] = gen(makesequence( gen("x^8-4*x^7+14*x^6-28*x^5+ 51*x^4-60*x^3+38*x^2-12*x+18",&ct), 
					   gen("1.2071067811865+2.0299824367188*i",&ct),
					   gen("-38/585*x^7+133/585*x^6-433/585*x^5+50/39*x^4-1303/585*x^3+1271/585*x^2-581/585*x+131/195",&ct),
					   gen("s^4+12*s^3+22*s^2-84*s-63",&ct), 
					   gen("-38/585*x^7+68/585*x^6-238/585*x^5+46/117*x^4-196/195*x^3+272/195*x^2+134/585*x+1/195",&ct),
					   gen("(-28*x^7+98*x^6-356*x^5+645*x^4-1268*x^3+1306*x^2-1204*x+111)/585",&ct),
					   gen("(-28*x^7+98*x^6-356*x^5+645*x^4-1268*x^3+1306*x^2-619*x+111)/585",&ct) ));

    (*minpols)["12s4c"] = gen(makesequence( gen("x^8-4*x^7+12*x^6-22*x^5+41*x^4-50*x^3+40*x^2-18*x+37",&ct), 
					    gen("1.3660254037844+1.82287565553228*i",&ct),
					    gen("-23/465*x^7+161/930*x^6-85/186*x^5+22/31*x^4-1277/930*x^3+668/465*x^2-893/930*x+707/930",&ct),
					    gen("s^4+ 8*s^3+4*s^2-48*s-48",&ct), 
					    gen("-23/465*x^7+34/465*x^6-73/465*x^5+17/155*x^4-313/465*x^3+389/465*x^2+158/465*x-13/93",&ct),
					    gen("(-74*x^7+259*x^6-805*x^5+1365*x^4-2863*x^3+3059*x^2-4267*x+268)/2790",&ct),
					    gen("(-74*x^7+259*x^6-805*x^5+1365*x^4-2863*x^3+3059*x^2-1477*x+268)/2790",&ct) ));

    (*minpols)["3s1"] = gen(makesequence( gen("x^4+4*x^3+4*x^2+3",&ct), 
					  gen("-2.22474487139160+.707106781186550*i",&ct),
					  gen("-1/2*x^3-3/2*x^2-1/2*x-1/2",&ct),
					  gen("s^2+4*s-2",&ct), 
					  gen("-1/2*x^3-3/2*x^2+1/2*x-1/2",&ct),
					  gen("-1/2*x^2-2*x-1/2",&ct),
					  gen("-1/2*x^2-x-1/2",&ct) ));

    /*        (*minpols)["4s1"] = gen(makesequence( gen("x^8+4*x^7+10*x^6+28*x^5+63*x^4+44*x^3-38*x^2-16*x+97",&ct), 
					  gen("-1.72474487139160+.841081377402110*i",&ct),
					  gen("-31/1694*x^7-181/2541*x^6-400/2541*x^5-727/1694*x^4-860/847*x^3-751/1694*x^2+5279/5082*x-3515/5082",&ct),
					  gen("s^4+4*s^3-82*s^2+212*s-71",&ct), 
					  gen("83/2541*x^7+232/2541*x^6+377/2541*x^5+494/847*x^4+1107/847*x^3+508/2541*x^2-1786/847*x+7199/2541",&ct),
					  gen("(204*x^7+111*x^6-349*x^5-817*x^4-4665*x^3-29621*x^2-57509*x+17000)/38962",&ct),
					  gen("(204*x^7+111*x^6-349*x^5-817*x^4-4665*x^3-29621*x^2-18547*x+17000)/38962",&ct) ));*/

    (*minpols)["4s1"] = gen(makesequence( gen("x^4-2*x^2+9",&ct), 
					  gen("i+sqrt(2)",&ct),
					  gen("-x^3/3+2*x/3",&ct),
					  gen("s^2-2",&ct), 
					  gen("-1/6*x^3+5/6*x",&ct),
					  gen("1/2*x^2-3/2",&ct),
					  gen("1/6*x^3+1/6*x",&ct) ));

    (*minpols)["6s1"] = gen(makesequence( gen("x^4-2*x^2+4",&ct), 
					  gen("-1.22474487139160+.707106781186550*i",&ct),
					  gen("-1/2*x^3+x",&ct),
					  gen("s^2-6",&ct),
					  gen("-1/2*x^3+2*x",&ct),
  					  gen("-1/2*x^2-x+1",&ct),
					  gen("-1/2*x^2+1",&ct) ));

    (*minpols)["2s5"] = gen(makesequence( gen("x^12+12*x^11+66*x^10+229*x^9+576*x^8+1116*x^7+1715*x^6+2136*x^5+2154*x^4+1748*x^3+1131*x^2+519*x+181",&ct), 
					  gen("-2.34680642174214-.431407536274166*i",&ct),
					  gen("-1/32*x^11-11/32*x^10-55/32*x^9-87/16*x^8-201/16*x^7-357/16*x^6-1001/32*x^5-1135/32*x^4-1019/32*x^3-729/32*x^2-201/16*x-149/32",&ct),
					  gen("s^6 + 12*s^5 + 48*s^4 + 73*s^3 + 42*s^2 + 6*s - 1",&ct), 
					  gen("-1/32*x^11 - 11/32*x^10 - 55/32*x^9 - 87/16*x^8 - 201/16*x^7 - 357/16*x^6 - 1001/32*x^5 - 1135/32*x^4 - 1019/32*x^3 - 729/32*x^2 - 185/16*x - 149/32",&ct),
					  gen("-1-x",&ct),
					  gen("-1",&ct) ));

    (*minpols)["3s5"] = gen(makesequence( gen("19-87*x+168*x^2-65*x^3-204*x^4+24*x^5+149*x^6+153*x^7+117*x^8+59*x^9+21*x^10+6*x^11+x^12",&ct),
					  gen("-1.84680642174214+.434617867510274*i",&ct),
					  gen("-178834/9168317*x^11-4254401/36673268*x^10-15019317/36673268*x^9-42776427/36673268*x^8-85983867/36673268*x^7-116971751/36673268*x^6-123189239/36673268*x^5-34991055/36673268*x^4+65864433/18336634*x^3+46519245/36673268*x^2-39465887/18336634*x+2092475/1930172",&ct),
					  gen("s^6 + 6*s^5 + 3*s^4 - 19*s^3 + 6*s^2 + 3*s - 1",&ct), 
					  gen("-178834/9168317*x^11 - 4254401/36673268*x^10 - 15019317/36673268*x^9 - 42776427/36673268*x^8 - 85983867/36673268*x^7 - 116971751/36673268*x^6 - 123189239/36673268*x^5 - 34991055/36673268*x^4 + 65864433/18336634*x^3 + 46519245/36673268*x^2 - 21129253/18336634*x + 2092475/1930172",&ct),
					  gen("-527574/9168317*x^11-3411894/9168317*x^10-12613710/9168317*x^9-36776259/9168317*x^8-78206085/9168317*x^7-115601734/9168317*x^6-130671429/9168317*x^5-73670769/9168317*x^4+71076527/9168317*x^3+57819627/9168317*x^2-78092519/9168317*x+984030/482543",&ct),
					  gen("-527574/9168317*x^11-3411894/9168317*x^10-12613710/9168317*x^9-36776259/9168317*x^8-78206085/9168317*x^7-115601734/9168317*x^6-130671429/9168317*x^5-73670769/9168317*x^4+71076527/9168317*x^3+57819627/9168317*x^2-68924202/9168317*x+984030/482543",&ct) ));

    (*minpols)["4s5"] = gen(makesequence( gen("x^24+12*x^22+18*x^21+66*x^20+108*x^19+371*x^18+162*x^17+594*x^16+342*x^15-492*x^14-5130*x^13+2257*x^12-3402*x^11-10272*x^10+12078*x^9+62064*x^8-59454*x^7-68964*x^6+74304*x^5+72261*x^4-147258*x^3+98577*x^2-33048*x+4869",&ct), 
					  gen(".754410156290088-.196187826424582*i",&ct),
					  gen("-150179616397372314156242190853/146043196431310993376218388837376*x^23 - 31171424878187911478032838659/48681065477103664458739462945792*x^22 - 5604379742760239208509926439857/438129589293932980128655166512128*x^21 - 3874453484714066030540135563649/146043196431310993376218388837376*x^20 - 12426011330195228008185824343037/146043196431310993376218388837376*x^19 - 24199873915635222822875075753425/146043196431310993376218388837376*x^18 - 11925065023198598194447420015027/24340532738551832229369731472896*x^17 - 2937692160117743700387747334283/6085133184637958057342432868224*x^16 - 206464824109625492897943813550795/219064794646966490064327583256064*x^15 - 17869198559252559328699768239953/18255399553913874172027298604672*x^14 - 536767168461560984639568714651/3042566592318979028671216434112*x^13 + 370604097781044195322077992717651/73021598215655496688109194418688*x^12 + 113338035292225506889977243303685/146043196431310993376218388837376*x^11 + 204030881037440195664433152051353/48681065477103664458739462945792*x^10 + 5817702218838097405612893559049051/438129589293932980128655166512128*x^9 - 560505962263492113155575493993957/146043196431310993376218388837376*x^8 - 9549653102809795067714264208510371/146043196431310993376218388837376*x^7 + 3007821583158537621187559695942007/146043196431310993376218388837376*x^6 + 3930499152149660117775679749905459/48681065477103664458739462945792*x^5 - 1272674176179768926381587526607421/48681065477103664458739462945792*x^4 - 6382404976484719486650424404062287/73021598215655496688109194418688*x^3 + 295600350700575739367349459350345/3042566592318979028671216434112*x^2 - 2143218245196344736723745649133255/48681065477103664458739462945792*x + 447564527073715219796701968368389/48681065477103664458739462945792",&ct),
					  gen("s^12+24*s^11+150*s^10-344*s^9-5337*s^8-4542*s^7+48596*s^6+61206*s^5-147705*s^4-148126*s^3+70167*s^2+39762*s-12851",&ct), 
					  gen("-134401075902646799941539078111/48681065477103664458739462945792*x^23 - 790688585160535099225140385819/438129589293932980128655166512128*x^22 - 15076330211623750848775485066421/438129589293932980128655166512128*x^21 - 10563297682852500389468260844077/146043196431310993376218388837376*x^20 - 33714008979524738660658231211037/146043196431310993376218388837376*x^19 - 66117413078563328521929400589329/146043196431310993376218388837376*x^18 - 97210650535102635063202672541207/73021598215655496688109194418688*x^17 - 146999950008028982932396136738195/109532397323483245032163791628032*x^16 - 565977385971334694854603378299661/219064794646966490064327583256064*x^15 - 33228941541837923792747402755461/12170266369275916114684865736448*x^14 - 21531855381993184507652494630667/36510799107827748344054597209344*x^13 + 990032077522405628313677687117773/73021598215655496688109194418688*x^12 + 358122647529567886422964308614873/146043196431310993376218388837376*x^11 + 4978191198780892211551265513411981/438129589293932980128655166512128*x^10 + 15830218925609839019705258115955979/438129589293932980128655166512128*x^9 - 1330734263028180538976665318236149/146043196431310993376218388837376*x^8 - 8543736380198992448030572331614949/48681065477103664458739462945792*x^7 + 2450599533002345148599803282459353/48681065477103664458739462945792*x^6 + 10582937733574322880168864485246979/48681065477103664458739462945792*x^5 - 9497393634553543842361383629706551/146043196431310993376218388837376*x^4 - 17138730167980857352748622149295601/73021598215655496688109194418688*x^3 + 3111281530761295588688257884210337/12170266369275916114684865736448*x^2 - 5459195774180115061269940661303803/48681065477103664458739462945792*x + 1097860674408400071146027751142865/48681065477103664458739462945792",&ct),
					  gen("-270123986295661903445952117649631/31107200839869241589134516822361088*x^23 - 71462314417601796065641167155795/10369066946623080529711505607453696*x^22 - 3393297362957474800587813945996853/31107200839869241589134516822361088*x^21 - 836748167377867000198244165192365/3456355648874360176570501869151232*x^20 - 7854907040544533721913658428221353/10369066946623080529711505607453696*x^19 - 15744436725086158656256486888517135/10369066946623080529711505607453696*x^18 - 67899058615374003830444945595652883/15553600419934620794567258411180544*x^17 - 6140208399777279722134623930551657/1296133368327885066213938200931712*x^16 - 133101326374560320148438263671016377/15553600419934620794567258411180544*x^15 - 8003446420883670338738920919549081/864088912218590044142625467287808*x^14 - 492541847899369270393607372993171/216022228054647511035656366821952*x^13 + 226753018805550905132622532440330925/5184533473311540264855752803726848*x^12 + 487406852241705623746244245792176979/31107200839869241589134516822361088*x^11 + 405596326204772132582963895690806029/10369066946623080529711505607453696*x^10 + 3655059954770134292756209470573722711/31107200839869241589134516822361088*x^9 - 53979072730916339288177794555091181/3456355648874360176570501869151232*x^8 - 5823159282541271227633916959081438687/10369066946623080529711505607453696*x^7 + 690017174240559140926450235981190913/10369066946623080529711505607453696*x^6 + 7145533838400376719262378065681236189/10369066946623080529711505607453696*x^5 - 287586336688956010482536637576889603/3456355648874360176570501869151232*x^4 - 3811814920474316984071068863220407215/5184533473311540264855752803726848*x^3 + 146809792661348715663393387257732733/216022228054647511035656366821952*x^2 - 940938690256441585546878669877996987/3456355648874360176570501869151232*x + 161980865452418573419770920854780391/3456355648874360176570501869151232",&ct), 
					  gen("-32396372046411839587126435188485/15553600419934620794567258411180544*x^23 - 2027767898712484951072772271979/1728177824437180088285250934575616*x^22 - 398132776344887088764578995419671/15553600419934620794567258411180544*x^21 - 268400180743626653572776147672637/5184533473311540264855752803726848*x^20 - 859115143865766487976404453135187/5184533473311540264855752803726848*x^19 - 1634421590383433426979622105218325/5184533473311540264855752803726848*x^18 - 7316663101003229779158764117608577/7776800209967310397283629205590272*x^17 - 182251451487796263140119273763505/216022228054647511035656366821952*x^16 - 12834026112737439582879179744858371/7776800209967310397283629205590272*x^15 - 1974550941288801615802158326768873/1296133368327885066213938200931712*x^14 + 40940907090669629158353803264959/108011114027323755517828183410976*x^13 + 29118910231028502589117095351558335/2592266736655770132427876401863424*x^12 + 32064682694389959760887918279490369/15553600419934620794567258411180544*x^11 + 14615920538602953779075064054497013/1728177824437180088285250934575616*x^10 + 406500616327210051935653996863227053/15553600419934620794567258411180544*x^9 - 56215486458398136034183665716469373/5184533473311540264855752803726848*x^8 - 709002771735407945384880081363164677/5184533473311540264855752803726848*x^7 + 232316218137670840651631291453366171/5184533473311540264855752803726848*x^6 + 882862150865459233436287981231974863/5184533473311540264855752803726848*x^5 - 96328307089364492298864138711758545/1728177824437180088285250934575616*x^4 - 473426130546026270721924665508287221/2592266736655770132427876401863424*x^3 + 21639030657990693451273219406174671/108011114027323755517828183410976*x^2 - 156945456077516364335771959277584953/1728177824437180088285250934575616*x + 31149707814430320934578485215132381/1728177824437180088285250934575616",&ct)
					  ));

    (*minpols)["4s5b"] = gen(makesequence( gen("x^8-28*x^6+238*x^4+372*x^2+441",&ct), 
					   gen("3.9681187850686669899+i",&ct),
					   gen("-13/4032*x^7+49/576*x^5-385/576*x^3-1409/1344*x",&ct),
					   gen("s^4-16*s^2+4",&ct), 
					   gen("-13/8064*x^7+343/8064*x^5-2695/8064*x^3-195/8064*x",&ct),
					   gen("-1/16128*x^7-1/256*x^6+1/2304*x^5+31/256*x^4+131/2304*x^3-283/256*x^2-5129/5376*x+381/256",&ct),
					   gen("13/8064*x^7-49/1152*x^5+385/1152*x^3+2753/2688*x",&ct) ));


  }

  return minpols;
}

