#include "FiniteGroup.h"
#include <cmath>
#include <iostream>
#include <fstream>
#include <string>

using namespace std;
using namespace giac;

FiniteGroup::FiniteGroup(string s): CMfield(s) {
  elements.push_back(Id);
  oldnumber=0;
};

bool FiniteGroup::areMultipleNum(gen M, gen N, gen eps) {
  gen Mn = matevaln(M);
  gen Nn = matevaln(N);
  gen A = Mn*inverse(Nn);
  //  cout << "A=" << A << endl;
  if (is_greater(abs(A[0][1],&ct),eps,&ct)) {
    //    cout << "01" << abs(A[0,1],&ct) << endl;
    return false;
  }
  else {
    if (is_greater(abs(A[0][2],&ct),eps,&ct)) {
      //      cout << "02" << abs(A[0,2],&ct) << endl;
      return false;
    }
    else {
      if (is_greater(abs(A[1][0],&ct),eps,&ct)) {
	//	cout << "10" << abs(A[1,0],&ct) << endl;
	return false;
      }
      else {
	if (is_greater(abs(A[1][2],&ct),eps,&ct)) {
	  //	  cout << "12" << abs(A[1,2],&ct) << endl;
	  return false;
	}
	else {
	  if (is_greater(abs(A[2][0],&ct),eps,&ct)) {
	    //	    cout << "20" << abs(A[2,0],&ct) << endl;
	    return false;
	  }
	  else {
	    if (is_greater(abs(A[2][1],&ct),eps,&ct)) {
	      //	      cout << "21" << abs(A[2,1],&ct) << endl;
	      return false;
	    }
	    else {
	      if (is_greater(abs(A[0][0]-A[1][1],&ct),eps,&ct)) {
		//		cout << "00-11" << abs(A[0,0]-A[1,1],&ct) << endl;
		return false;
	      }
	      else {
		if (is_greater(abs(A[1][1]-A[2][2],&ct),eps,&ct)) {
		  //		  cout << "11-22" << abs(A[1,1]-A[2,2],&ct) << endl;
		  return false;
		}
		else {
		  return true;
		}
	      }
	    }
	  }			      
	}
      }
    }
  }		      
}

bool FiniteGroup::areMultiple(gen M, gen N) {
  // starts by finding a nonzero element in M, then uses it as a "pivot"
  bool fdnz = false;
  int k=-1;
  int l;
  while (!fdnz && k!=2) {
    k++;
    l=-1; 
    while (!fdnz && l!=2) {
      l++;
      fdnz = !operator_equal(M[k][l],zeronumber,&ct);
      if (!fdnz && !operator_equal(N[k][l],zeronumber,&ct)) {
	return false;
      }
    }
  }
  if (!fdnz) {
    cerr << "Matrix is zero, this should not happen in a group!" << endl;
    exit(1);
  }
  else {
    gen a = M[k][l];
    gen b = N[k][l];
    bool ismult = true;
    while (ismult && l<3) {
      gen te = b*M[k][l]-a*N[k][l];
      red(te);
      ismult = operator_equal(te,zeronumber,&ct);
      l++;
    }
    k++;
    while (ismult && k<3) {
      l=0;
      while (ismult && l<3) {
	gen te = b*M[k][l]-a*N[k][l];
	red(te);
	ismult = operator_equal(te,zeronumber,&ct);
	l++;
      }
      k++;
    }
    return ismult;
  }
}

/*
bool FiniteGroup::isNew(gen M) {
  return !isInList(M,elements);
}
*/

bool FiniteGroup::isInList(gen M, vector<gen> list) {
  bool res = false;
  int k=0;
  while (!res && k<list.size()) {
    //    res = areMultiple(M,list[k]);
    res = areMultipleNum(M,list[k],1e-6);
    k++;
  }
  return res;
}

void FiniteGroup::addElement(gen M) {
  if (!isInList(M,elements)) {
    //    cout << "Added new elt: " << M << endl;
    elements.push_back(M);
  }
}

void FiniteGroup::addElements(vector<gen> list) {
  for (unsigned k=0; k<list.size(); k++) {
    addElement(list[k]);
  }  
}

void FiniteGroup::expand() {
  vector<gen> newlist;
  for (unsigned k=oldnumber; k<elements.size(); k++) {
    for (unsigned l=1; l<elements.size(); l++) {
      gen M = elements[k]*elements[l];
      matred(M);
      if (!isInList(M,elements) && !isInList(M,newlist)) {
	//	cout << "New matrix: " << M << endl;
	newlist.push_back(M);
      }
    }
  }
  for (unsigned l=oldnumber; l<elements.size(); l++) {
    for (unsigned k=1; k<elements.size(); k++) {
      gen M = elements[k]*elements[l];
      matred(M);
      if (!isInList(M,elements) && !isInList(M,newlist)) {
	//	cout << "New matrix: " << M << endl;
	newlist.push_back(M);
      }
    }
  }
  oldnumber = elements.size();
  //  cout << "old nb: " << oldnumber << endl;
  addElements(newlist);
  cout << "Current nb of elements: " << elements.size() << endl;
}

void FiniteGroup::expandToGroup() {
  bool done = false;
  int count = 0;
  while (!done && count<20) {
    int ne = elements.size();
    //    cout << "ne=" << ne << endl;
    expand();
    done = (ne==elements.size());
    count++;
  }
  if (done) {
    cout << "Found a group with " << elements.size() << " elements" << endl;
  }
  else {
    cout << "Did not find a group, even after expanding 20 times..." << endl;    
  }
}
