#include "FiniteLinearGroup.h"
#include <cmath>
#include <iostream>
#include <fstream>
#include <string>

using namespace std;
using namespace giac;

FiniteLinearGroup::FiniteLinearGroup(string s): CMfield(s) {
  elements.push_back(Id2);
  oldnumber=0;
};

bool FiniteLinearGroup::isInList(gen M, vector<gen> list) {
  bool res = false;
  int k=0;
  while (!res && k<list.size()) {
    res = operator_equal(M,list[k],&ct);
    k++;
  }
  return res;
}

void FiniteLinearGroup::addElement(gen M) {
  if (!isInList(M,elements)) {
    elements.push_back(M);
  }
}

void FiniteLinearGroup::addElements(vector<gen> list) {
  for (unsigned k=0; k<list.size(); k++) {
    addElement(list[k]);
  }  
}

void FiniteLinearGroup::expand() {
  //  cout << "expand..." << endl;
  vector<gen> newlist;
  for (unsigned k=oldnumber; k<elements.size(); k++) {
    for (unsigned l=1; l<elements.size(); l++) {
      gen M = elements[k]*elements[l];
      matred(M);
      if (!isInList(M,elements) && !isInList(M,newlist)) {
	//	cout << "New matrix: " << M << endl;
	newlist.push_back(M);
      }
    }
  }
  for (unsigned l=oldnumber; l<elements.size(); l++) {
    for (unsigned k=1; k<elements.size(); k++) {
      gen M = elements[k]*elements[l];
      matred(M);
      if (!isInList(M,elements) && !isInList(M,newlist)) {
	//	cout << "New matrix: " << M << endl;
	newlist.push_back(M);
      }
    }
  }
  oldnumber = elements.size();
  //  cout << "old nb: " << oldnumber << endl;
  addElements(newlist);
  cout << "Current nb of elements: " << elements.size() << endl;
}

void FiniteLinearGroup::expandToGroup() {
  bool done = false;
  int count = 0;
  while (!done && count<20) {
    int ne = elements.size();
    //    cout << "ne=" << ne << endl;
    expand();
    done = (ne==elements.size());
    count++;
  }
  if (done) {
    cout << "Found a group with " << elements.size() << " elements" << endl;
  }
  else {
    cout << "Did not find a group, even after expanding 20 times..." << endl;    
  }
}
