
#ifndef POLYTOPENOJ_H
#define POLYTOPENOJ_H
 
#include <iostream>
#include <giac/giac.h>

#include <string>
#include <map>
#include <algorithm>

using namespace std;
using namespace giac;

#include "CMfield.h"
#include "Prism.h"
#include "Util.h"
#include "edge.h"
#include "ridge.h"
#include "sgraph.h"
#include "FiniteGroup.h"
#include "FiniteLinearGroup.h"

extern context ct;
class PolytopeNoJ: public CMfield
{
  public:

  PolytopeNoJ(string s);

  bool emb_dim1, emb_dim2, check_cycles, check_euler, check_sphere, perform_algo, draw_pics;

  gen p0, p0c;
 
  gen H; // Matrix of the ambient Hermitian form
  gen R1, R2, R3, R1i, R2i, R3i, J, Ji, P, Pi, Ppower;

  gen Q; // change of basis to diagonal form, using eigenvectors of P

  // Matrix of the symmetry of the polytope (this may be (123) or (1J))
  bool issymmetric;
  
  gen mult, multconj;

  WVector e1w, e2w, e3w;

  gen extramirror;

  int orderofp, orderofppower;

  gen euler_char;

  vector<int> word_p1, word_m1, word_123, word_321;
  vector<Prism> faces;
  vector<Prism> faceReps;

  vector<WVector> vertices;
  vector<WVector> vertReps;

  vector<vector<int> > vertexPOrbits;
  vector<vector<int> > vertexOrbits;
  vector<int> currentVertexOrbit;
  vector<int> orbitsToVisit;

  vector<vector<int> > facePOrbits;
  vector<vector<int> > edgePOrbits;
  vector<vector<int> > ridgePOrbits;

  vector<vector<int> > vertexPairings;
  vector<vector<int> > vertexPOrbitPairings;
  vector<vector<int> > vertexPOrbitPairingMaps;
  vector<vector<int> > edgePairings;
  vector<vector<int> > ridgePairings;
  vector<int> facePairings;

  vector<edge> edges;
  vector<ridge> ridges;

  vector<vector<int> > relations;

  gen Proj(gen v, gen w);  
  gen Inn(gen v, gen w);
  gen SqNorm(gen v);
  gen BoxProd(gen v, gen w);

  gen refl(gen x, gen v, gen z);
  gen reflMat(gen v, gen z);

  bool testSame(WVector V, WVector W);
  bool isInList(WVector V, vector<WVector> li);
  int findInList(WVector V, vector<WVector> li);
  int findInList(gen V, vector<WVector> li);

  void initSporadicGroup();
  //  void initThompsonGroup(string s); // not implemented yet!
  //  void initMostowGroup(string s);   // not implemented yet!

  Prism createPrism(WVector v0, WVector v1, WVector v2);
  void updateFreeRidges();
  void expand();

  bool isPrismNew(Prism A);
  bool isVertexNew(WVector V);
  bool comparePrisms(Prism A, Prism B);
  bool samePolygon(vector<WVector> V, vector<WVector> W);  
  bool samePolygon(vector<gen> V, vector<gen> W);  

  bool computeVertices(Prism &A);

  bool isOnPrism(gen X, Prism A);
  bool isOnMirror(gen X, Prism A);
  bool isOnTopFace(gen X, Prism A);
  gen reflectAcrossSpine(gen X, Prism A);

  Prism createPimage(Prism A);
  Prism createPinvimage(Prism A);

  Prism applyReflection(Prism A);
  Prism applyInverseReflection(Prism A);

  Prism applyMatrix(gen M, Prism A);

  void addFaceOrbit(Prism A);

  bool isVertexSafe(WVector V, Prism A);

  vector<int> findFaces(WVector v);
  vector<int> findFaces(gen V);
  vector<int> findBisectors(WVector v);
  vector<int> findBisectors(gen V);
  vector<WVector> findMirrors(WVector v);
  vector<WVector> findMirrors(gen V);

  void useRelations(vector<int> &w);
  void useRelations(Prism &A);

  void computeStabilizer(int k);
  void studyVertexStabilizers();
  void findNextVertexInCycle(unsigned k);

  void addEdge(edge e);
  bool isSame(edge e1, edge e2);
  bool isFlipped(edge e1, edge e2);
  bool isEdgeNew(edge e);

  void addRidge(ridge e);
  bool isSame(ridge e1, ridge e2);
  bool isRotated(ridge e1, ridge e2);
  bool isRidgeNew(ridge e);

  int applyPtoVertex(int k);
  vector<int> getPOrbitOfVertex(int k);
  vector<int> getPOrbitOfFace(int k);
  int findNextInFacePOrbit(int k);
  void addEdgeOld(edge e);

  void createBraidRelation(vector<int> w1, vector<int> w2);

  gen evalWord(vector<int> w);

  int applyPairing(ridge e);
  void checkCycles();

  int applyPairingToVertex(int jface, int j);
  int applyPairingToEdge(int jface, int j);
  int applyPairingToRidge(int jface, int j);
  
  void drawFace(unsigned j); 
  void drawProjection(unsigned axis); 

  bool checkSide();
  void checkEdge(unsigned j);
  void checkRidge(unsigned j);

  void checkPairings();
  void computeAllVertices();
  void computeAllEdges();
  void computeAllRidges();

  bool checkt1val(gen rur, gen par, gen t1v, int &res);
  bool checkt1valAttempt(gen rur, gen par, gen t1v, int &res);
  bool checkEqualt1val(gen rur, gen par, gen t1v);

  bool checkt2val(gen rur, gen par, gen t1v, int &res);
  bool checkt2valAttempt(gen rur, gen par, gen t1v, int &res);
  bool checkEqualt2val(gen rur, gen par, gen t1v);

  bool checkt3val(gen rur, gen par, gen t3v, gen vl, gen wl, gen vt, int &res);
  bool checkt3valAttempt(gen rur, gen par, gen t3v, gen vl, gen wl, gen vt, int &res);
  bool checkEqualt3val(gen rur, gen par, gen t3v, gen vl, gen wl, gen vt);

  bool checkEqSign(gen rur, gen par, gen eq, int &res);
  bool checkEqSignAttempt(gen rur, gen par, gen eq, int &res);
  bool checkEqZero(gen rur, gen par, gen eq);

  void drawRidgePic(unsigned i);

  void computeEulerCharacteristic();
  gen eulerdim0();
  gen eulerdim1();
  gen eulerdim2();
  gen eulerdim3();

  bool isStabilizedByPpower(int j);
  bool isRidgeComplex(int j, gen &mi);

  void exportFaceData();

  void checkSphere();
  void checkLinks();

  gen computeAngle(int j, gen X);
  gen computeAngleNum(int j, gen X);

  bool expressAsWord(gen M, vecteur list, gen fixed_pt, gen mirror, vector<int> &inds);
  
  bool isArithmetic();
  string checkSignature(gen A);

};

#endif
