#include "PolytopeSmall.h"
#include <cmath>
#include <iostream>
#include <fstream>
#include <string>

using namespace std;
using namespace giac;

PolytopeSmall::PolytopeSmall(string s): CMfield(s) {
  importSporadicGroup();
};

void PolytopeSmall::importSporadicGroup() {

  cout << "Initializing sporadic group..." << endl;

  word_p1.push_back(1);
  word_m1.push_back(-1);

  vector<int> w1;
  w1.push_back(1);
  vector<int> w2;
  w2.push_back(2);
  vector<int> w3;
  w3.push_back(3);

  mult = v;
  multconj = myconj(v);
 
  gen tauc = myconj(tau);
  gen tee = tau*tauc;
  red(tee);

  gen ts;
  gen de = _degree(makesequence(tee,x__IDNT_e),&ct);
  cout << "ts^3=" << tee << endl;
  if (operator_equal(de,zeronumber,&ct)) {
    gen eq = gen("x^3",&ct);
    eq = eq - tee;
    ts = _solve(eq,&ct)[0];
    if (ts.type!=0) {
      cerr << "Trouble, this sporadic group does not seem to satisfy |tau|^2 in Q" << endl;
      exit(1);
    }
  }
  else {
    cerr << "Trouble, this sporadic group does not seem to satisfy |tau|^2 in Q" << endl;
    exit(1);
  }
  cout << "|tau|^2=" << ts << endl;

  gen a = 2-mult-multconj;
  gen b12 = (multconj-1)*tau;
  gen b23 = (multconj-1)*tauc*ts;
  gen b13 = (1-multconj)*tauc;
  
  vecteur coh(9);
  coh[0] = a;           coh[1] = b12;         coh[2] = b13;
  coh[3] = myconj(b12); coh[4] = a*ts*ts;           coh[5] = b23;
  coh[6] = myconj(b13); coh[7] = myconj(b23); coh[8] = a*ts*ts;
  H = createMatrix(coh);

  vecteur cor1(9);
  cor1[0] = mult;  cor1[1] = tau;  cor1[2] = -tauc;
  cor1[3] = 0;  cor1[4] = 1;    cor1[5] = 0;
  cor1[6] = 0;  cor1[7] = 0;    cor1[8] = 1;
  R1 = createMatrix(cor1);
  matred(R1);

  vecteur cor2(9);
  cor2[0] = 1;  cor2[1] = 0;  cor2[2] = 0;
  cor2[3] = -(tauc*mult/ts)/ts;  cor2[4] = mult;    cor2[5] = tauc/ts;
  cor2[6] = 0;  cor2[7] = 0;    cor2[8] = 1;
  R2 = createMatrix(cor2);
  matred(R2);

  vecteur cor3(9);
  cor3[0] = 1;  cor3[1] = 0;    cor3[2] = 0;
  cor3[3] = 0;  cor3[4] = 1;    cor3[5] = 0;
  cor3[6] = (tau*mult/ts)/ts;  cor3[7] = -tau*mult/ts;    cor3[8] = mult;
  R3 = createMatrix(cor3);
  matred(R3);
  
  vecteur cor1i(9);
  cor1i[0] = multconj;  cor1i[1] = -tau*multconj;  cor1i[2] = tauc*multconj;
  cor1i[3] = 0;  cor1i[4] = 1;    cor1i[5] = 0;
  cor1i[6] = 0;  cor1i[7] = 0;    cor1i[8] = 1;
  R1i = createMatrix(cor1i);
  matred(R1i);

  vecteur cor2i(9);
  cor2i[0] = 1;  cor2i[1] = 0;  cor2i[2] = 0;
  cor2i[3] = (tauc/ts)/ts;  cor2i[4] = multconj;    cor2i[5] = -tauc*multconj/ts;
  cor2i[6] = 0;  cor2i[7] = 0;    cor2i[8] = 1;
  R2i = createMatrix(cor2i);
  matred(R2i);

  vecteur cor3i(9);
  cor3i[0] = 1;  cor3i[1] = 0;    cor3i[2] = 0;
  cor3i[3] = 0;  cor3i[4] = 1;    cor3i[5] = 0;
  cor3i[6] = -(tau/ts)/ts;  cor3i[7] = tau/ts;    cor3i[8] = multconj;
  R3i = createMatrix(cor3i);
  matred(R3i);

  vecteur coq;
  // this is done in a strange place, should define a global variable...
  if (tautag=="s4c") {
    coq.push_back(gen("exp(-2*pi*i/7)",&ct));
    coq.push_back(gen("exp(-8*pi*i/7)",&ct));
    coq.push_back(gen("exp(10*pi*i/7)",&ct));
    coq.push_back(onenumber);
    coq.push_back(onenumber);
    coq.push_back(onenumber);
    coq.push_back(gen("exp(2*pi*i/7)",&ct));
    coq.push_back(gen("exp(8*pi*i/7)",&ct));
    coq.push_back(gen("exp(-10*pi*i/7)",&ct));
  }
  else {
    if (tautag=="s1") {
      coq.push_back(gen("exp(pi*i/3)",&ct));
      coq.push_back(gen("exp(pi*i/12)",&ct));
      coq.push_back(gen("exp(-5*pi*i/12)",&ct));
      coq.push_back(onenumber);
      coq.push_back(onenumber);
      coq.push_back(onenumber);
      coq.push_back(gen("exp(-pi*i/3)",&ct));
      coq.push_back(gen("exp(-pi*i/12)",&ct));
      coq.push_back(gen("exp(5*pi*i/12)",&ct));
    }
    else {
      if (tautag=="s5") {
        coq.push_back(gen("exp(13*pi*i/45)",&ct));
        coq.push_back(gen("exp(2*pi*i/9)",&ct));
        coq.push_back(gen("exp(-23*pi*i/45)",&ct));
        coq.push_back(onenumber);
        coq.push_back(onenumber);
        coq.push_back(onenumber);
        coq.push_back(gen("exp(-13*pi*i/45)",&ct));
        coq.push_back(gen("exp(-2*pi*i/9)",&ct));
        coq.push_back(gen("exp(23*pi*i/45)",&ct));
      }
      else {
        cerr << "Unknown tau value.." << endl;
        exit(1);
      }
    }
  }
  gen Q = createMatrix(coq);
  //  cout << "Q=" << Q << endl;

  vecteur E;
  E.push_back(convert_interval(onenumber,20,&ct));
  gen old_tau = gen("exp(-pi*i/9)*(i*sqrt(3)+sqrt(5))/2",&ct);
  gen old_tauc = gen("exp(pi*i/9)*(-i*sqrt(3)+sqrt(5))/2",&ct);
  gen tee1 = _simplify(onenumber/(old_tau*old_tau),&ct);
  gen tee2 = _simplify(onenumber/(old_tauc*old_tauc),&ct);
  E.push_back(_evalf(makesequence(tee1,20),&ct));
  E.push_back(_evalf(makesequence(tee2,20),&ct));

  cout << "E: " << E << endl;

  vecteur p0temp;
  cout << "Creating p0" << endl;
  for (unsigned k=0; k<3; k++) {
    p0temp.push_back(E[k]*convert_interval(Q[k][0],20,&ct));
  }
  p0 = gen(p0temp);
  p0c = conj(p0,&ct);
  
  cout << "p0=" << p0 << endl;

  gen R123 = R1*R2*R3;
  matred(R123);

  gen Xte1 = evali(R123)*p0;
  
  cout << Xte1/Xte1[1] << endl;
  cout << p0/p0[1] << endl;
  cout << "Sq nm of p0: " << p0c*evali(H)*p0 << endl;

  
  int o123 = order(R123);
  vector<int> w123;
  for (unsigned k=0; k<o123; k++) {
    w123.push_back(1);
    w123.push_back(2);
    w123.push_back(3);
  }

  relations.push_back(w123);
  relations.push_back(flipword(w123));

  vector<int> w232b;
  w232b.push_back(2);
  w232b.push_back(3);
  w232b.push_back(-2);

  vector<int> w3b23;
  w3b23.push_back(-3);
  w3b23.push_back(2);
  w3b23.push_back(3);

  createBraidRelation(word_p1,w2);
  createBraidRelation(word_p1,w232b);
  createBraidRelation(word_p1,w3b23);

  cout << "Relations:" << endl;
  for (unsigned k=0; k<relations.size(); k++) {
    cout << "  " << convertword(relations[k]) << endl;
  }

  cout << "Check matrices..." << endl;

  cout << "R1 check: " << endl;
  gen te = transpose(matconj(R1))*H*R1-H;
  matred(te);
  cout << te << endl;
 
  cout << "R2 check: " << endl;
  gen te2 = _tran(matconj(R2),&ct)*H*R2-H;
  matred(te2);
  cout << te2 << endl;
  
  cout << "R3 check: " << endl;
  gen te4 = _tran(matconj(R3),&ct)*H*R3-H;
  matred(te4);
  cout << te4 << endl;
  
  e1w = WVector(e1,w1);
  e2w = WVector(e2,w2);
  e3w = WVector(e3,w3);

  importFaceData();

  bool test = checkSide();
  if (test) {
    cout << "The fixed point of P is on the correct side of all faces" << endl;
  }
  else {
    cout << "The fixed point of P does not seem to be on the correct side (increasing precision may help)..." << endl;
  }

  //  checkRidge(0,30);

}

gen PolytopeSmall::Proj(gen v, gen w) {
  gen fa = inverse(Inn(w,w));
  if (operator_equal(fa,zeronumber,&ct)) {
    cerr << "Do you really mean to divide by zero?" << endl;
    exit(1);
  }
  else {
    gen te = _simplify(v-Inn(v,w)*fa*w,&ct);
    //    cout << "te=" << te << endl;
    vectred(te);  
    //    cout << "   reduced... " << te << endl;
    return te;
  }
}

gen PolytopeSmall::Inn(gen v, gen w) {
  gen res = vectconj(w)*H*v;
  red(res);
  return res;
}

gen PolytopeSmall::SqNorm(gen v) {
  gen res = vectconj(v)*H*v;
  red(res);
  return convertToReal(res);
}

gen PolytopeSmall::BoxProd(gen v, gen w) {
  gen X = vectconj(v)*H;
  gen Y = vectconj(w)*H;
  vecteur res(3);
  res[0] = X[1]*Y[2] - X[2]*Y[1];
  res[1] = X[2]*Y[0] - X[0]*Y[2];
  res[2] = X[0]*Y[1] - X[1]*Y[0];
  gen bp = gen(res);
  vectred(bp);
  return bp;
}

gen PolytopeSmall::refl(gen X, gen V, gen z) {
  gen te = Inn(V,V);
  if (te==zeronumber) {
    cerr << "Trying to reflect across a null vector!" << endl;
    exit(1);
  }
  else {
    gen yo = _simplify(X + (z-1)*Inn(X,V)*inverse(te)*V,&ct);
    vectred(yo);
    return yo;
  }
}

gen PolytopeSmall::reflMat(gen u, gen z) {
  vecteur te;
  gen te1 = refl(e1,u,z);
  gen te2 = refl(e2,u,z);
  gen te3 = refl(e3,u,z);
  for (unsigned k=0; k<3; k++) {
    te.push_back(te1[k]);
    te.push_back(te2[k]);
    te.push_back(te3[k]);
  }
  return createMatrix(te);
}

Prism PolytopeSmall::createPrism(WVector v0, WVector v1, WVector v2){  

  Prism A;
  
  A.jstart = 0;

  A.mirror = v0;

  A.sides.push_back(v1); 
  A.sides.push_back(v2); 

  vector<int> w; // provisoire, on trouvera w plus tard
  A.apex = WVector(BoxProd(v1.coords,v2.coords),w);

  A.apexProjection = Proj(A.apex.coords,A.mirror.coords);
  //  vectred(A.apexProjection);
  A.cspine = BoxProd(A.apex.coords,A.apexProjection);

  /*  cout << "Apex: " << A.apex << " sq norm: " << SqNorm(A.apex) << endl;
  cout << "Apex projection: " << A.apexProjection <<  " sq norm: " << evali(SqNorm(A.apexProjection)) << endl;
  cout << "Polar to complex spine: " << A.cspine <<  " sq norm: " << evali(SqNorm(A.cspine)) << endl;*/
 
  bool done = false;
  unsigned n = 2;
  while (!done && n<100) {
    //    cout << "TEST n=" << n << endl;
    gen M1 = reflMat(A.sides[0].coords,mult);
    vector<int> wl = conjugateword(A.sides[0].word,A.sides[1].word); 
    //   cout << "wl: " << wl << endl;
    gen vl = M1*A.sides[1].coords; 
    vectred(vl);
    WVector Vl(vl,wl);
    if (!testSame(Vl,A.sides[n-1])) {
      A.sides.insert(A.sides.begin(),Vl);
      n++;
      A.jstart++;
    }
    else {
      done = true;
    }
    if (!done) {
      gen M2 = reflMat(A.sides[n-1].coords,multconj);
      vector<int> wr = conjugateword(invertword(A.sides[n-1].word),A.sides[n-2].word);
      gen vr = M2*A.sides[n-2].coords;
      vectred(vr);
      WVector Vr(vr,wr);
      //      cout << "TEST vr=" << vr << endl;
      if (!testSame(Vr,A.sides[0])) {
	A.sides.push_back(Vr);
	n++;
      }
      else {
	done = true;
      }
    }
  }
  if (!done) {
    cerr << "Braiding cycle did not close up!" << endl;
    exit(1);
  }
  else {

    unsigned k1 = 0;
    unsigned l1 = A.sides[0].word.size();
    for (unsigned k=1; k<A.sides.size(); k++) {
      unsigned l = A.sides[k].word.size();
      if (l<l1) {
	l1 = l;
	k1 = k;
      }
    }
    unsigned k2;
    if (k1==0) {
      if (A.sides[A.sides.size()-1].word.size()<A.sides[1].word.size()) {
	k2 = A.sides.size()-1;
      }
      else {
	k2 = 1;
      }
    }
    else {
      k2 = (k1+1)%A.sides.size();
    }

    gen M1 = reflMat(A.sides[k1].coords,mult)*reflMat(A.sides[k2].coords,mult);
    matred(M1);
    int om1 = order(M1);
    cout << "Product of order " << om1 << endl;
    int bo = braidOrder(reflMat(A.sides[k1].coords,mult),reflMat(A.sides[k2].coords,mult));
    cout << "Braiding order " << bo << endl;

    vector<int> w0 = concat(A.sides[k1].word,A.sides[k2].word);
    vector<int> w = concat(A.sides[k1].word,A.sides[k2].word);

    bool isrefl = false;
    unsigned k=1;
    gen M = M1;
    while (k<om1 && !isrefl) {
      gen dt = discrAdjusted(M);
      isrefl = operator_equal(dt,zeronumber,&ct); // THIS IT NOT QUITE THE RIGHT TEST (COULD BE PARABOLIC)
      if (isrefl) {
	int sg;
	if (!checkSign(SqNorm(A.apex.coords),sg)) {
	  cerr << "Problem checking sign" << endl;
	  exit(1);
	}
	else {
	  if (sg>0) {
	    gen X1 = BoxProd(A.cspine,A.apex.coords);
	    gen X2 = M*X1;
	    vectred(X2);
	    isrefl = areDep(X1,X2);
	  }
	  else {
	    if (sg<0) {
	      gen X1 = A.apex.coords;
	      gen X2 = M*X1;
	      vectred(X2);
	      isrefl = areDep(X1,X2);
	    }
	    else {
	      // don't know what to do...
	    }
	  }
	}
      }
      if (isrefl) {
	cout << "Found reflection, power " << k << endl;
      }
      else {
	M = M*M1;
	w = concat(w,w0);
	matred(M);
      }
      k++;
    }
    if (!isrefl) {
      cerr << "Did not find any reflection power..." << endl;
      //      exit(1);
    }

    useRelations(w);
    A.apex.word = w;
    cout << "word for apex=" << convertword(w) << endl;

    unsigned ns = A.sides.size();
    cout << "Constructed a pyramid with " << ns << " sides!" << endl;
    A.isSurrounded = false;
    for (unsigned k=0; k<ns; k++) {
      A.freeRidges.push_back(k);
      //      cout << A.sides[k].word << endl;
    }
    //    cout << "jstart: " << A.jstart << endl;
    return A;
  }

}

bool PolytopeSmall::isPrismNew(Prism A){
  unsigned k=0;
  bool res = true;
  while (res && k<faces.size()) {
    res = !comparePrisms(A,faces[k]);
    k++;
  }
  return res;
}

bool PolytopeSmall::isVertexNew(WVector V) {
  return !isInList(V,vertices);
}

bool PolytopeSmall::comparePrisms(Prism A, Prism B) {
  unsigned ns = A.sides.size();
  if (!testSame(A.mirror,B.mirror) || ns!=B.sides.size()) {
    //  if (!areDep(A.mirror.coords,B.mirror.coords) || ns!=B.sides.size()) {
    return false;
  }
  else {
    return samePolygon(A.sides,B.sides);
  }
}

bool PolytopeSmall::samePolygon(vector<WVector> V, vector<WVector> W){
  unsigned vs = V.size();
  if (W.size()!=vs) {
    return false;
  }
  else {
    bool foundsame = false;
    unsigned k=0;
    while (!foundsame && k<vs) {
      foundsame = testSame(V[0],W[k]);
      k++;
    }
    if (!foundsame) {
      return false;
    }
    else {
      //      cout << "Found common vector..." << endl;
      unsigned l=1;
      bool same = true;
      while (same && l<V.size()) {
	k = k%vs;
	same = testSame(V[l],W[k]);
	k++;
	l++;
      }
      return same;
    }
  }
}

Prism PolytopeSmall::applyReflection(Prism A) {

  gen M = reflMat(A.mirror.coords,mult);
  vector<int> w = A.mirror.word;

  Prism A1;
 
  gen m = M*A.mirror.coords;
  vectred(m);
  vector<int> w1 = conjugateword(w,A.mirror.word);
  A1.mirror = WVector(m,w1);

  A1.apex = WVector(M*A.apex.coords,conjugateword(w,A.apex.word));
  vectred(A1.apex.coords);

  A1.apexProjection = M*A.apexProjection;
  vectred(A1.apexProjection);

  A1.cspine = M*A.cspine;
  vectred(A1.cspine);

  // I think we want to switch, here, right?
  /*  A1.X0 = M*A.X1;
  A1.X1 = M*A.X0;
  vectred(A1.X0);
  vectred(A1.X1);*/

  for (unsigned k=0; k<A.sides.size(); k++) {
      vector<int> wk = conjugateword(w,A.sides[k].word);
      gen sk = M*A.sides[k].coords;
      vectred(sk);
      A1.sides.push_back(WVector(sk,wk));
  }

  for (unsigned k=0; k<A.bottomface.size(); k++) {
      vector<int> wk = conjugateword(w,A.bottomface[k].word);
      gen sk = M*A.bottomface[k].coords;
      vectred(sk);
      A1.bottomface.push_back(WVector(sk,wk));
  }
  for (unsigned k=0; k<A.topface.size(); k++) {
      vector<int> wk = conjugateword(w,A.topface[k].word);
      gen sk = M*A.topface[k].coords;
      vectred(sk);
      A1.topface.push_back(WVector(sk,wk));
  }
  for (unsigned k=0; k<A.midvertices.size(); k++) {
      vector<int> wk = conjugateword(w,A.midvertices[k].word);
      gen sk = M*A.midvertices[k].coords;
      vectred(sk);
      A1.midvertices.push_back(WVector(sk,wk));
      A1.midvertindices.push_back(A.midvertindices[k]);
  }

  A1.isSurrounded = false;
  for (unsigned k=0; k<A1.sides.size(); k++) {
    A1.freeRidges.push_back(k);
  }
  A1.jstart = A.jstart;
 
  return A1;
  
}

Prism PolytopeSmall::applyInverseReflection(Prism A) {

  gen M = reflMat(A.mirror.coords,multconj);
  vector<int> w = A.mirror.word;

  Prism A1;
 
  gen m = M*A.mirror.coords;
  vectred(m);
  vector<int> w1 = conjugateword(w,A.mirror.word);
  A1.mirror = WVector(m,w1);

  A1.apex = WVector(M*A.apex.coords,conjugateword(w,A.apex.word));
  vectred(A1.apex.coords);
  A1.apexProjection = M*A.apexProjection;
  vectred(A1.apexProjection);
  A1.cspine = M*A.cspine;
  vectred(A1.cspine);
  
  // I think we want to switch, here, right?
  /*  A1.X0 = M*A.X1;
  A1.X1 = M*A.X0;
  vectred(A1.X0);
  vectred(A1.X1);*/

  for (unsigned k=0; k<A.sides.size(); k++) {
      vector<int> wk = conjugateword(w,A.sides[k].word);
      gen sk = M*A.sides[k].coords;
      vectred(sk);
      A1.sides.push_back(WVector(sk,wk));
  }

  for (unsigned k=0; k<A.bottomface.size(); k++) {
      vector<int> wk = conjugateword(w,A.bottomface[k].word);
      gen sk = M*A.bottomface[k].coords;
      vectred(sk);
      A1.bottomface.push_back(WVector(sk,wk));
  }
  for (unsigned k=0; k<A.topface.size(); k++) {
      vector<int> wk = conjugateword(w,A.topface[k].word);
      gen sk = M*A.topface[k].coords;
      vectred(sk);
      A1.topface.push_back(WVector(sk,wk));
  }
  for (unsigned k=0; k<A.midvertices.size(); k++) {
      vector<int> wk = conjugateword(w,A.midvertices[k].word);
      gen sk = M*A.midvertices[k].coords;
      vectred(sk);
      A1.midvertices.push_back(WVector(sk,wk));
      A1.midvertindices.push_back(A.midvertindices[k]);
  }

  A1.isSurrounded = false;
  for (unsigned k=0; k<A1.sides.size(); k++) {
    A1.freeRidges.push_back(k);
  }
  A1.jstart = A.jstart;

  return A1;
  
}

bool PolytopeSmall::isRidgeComplex(int j, gen &mi) {
  ridge r = ridges[j];
  unsigned jface = r.faceindices[0];
  
  bool iscomplex = false;
  bool isbottom = true;
  int uu=0;
  while (isbottom && uu<r.verts.size()) {
    gen X = vertices[r.verts[uu]].coords;
    isbottom = operator_equal(Inn(X,faces[jface].mirror.coords),zeronumber,&ct);
    uu++;
  }
  if (isbottom) {
    //    cout << "This is a complex ridge!" << endl;
    mi = faces[jface].mirror.coords;
    return true;
  }
  else {
    bool istop = true;
    uu=0;
    while (istop && uu<r.verts.size()) {
      gen X = vertices[r.verts[uu]].coords;
      istop = operator_equal(Inn(X,faces[jface].apex.coords),zeronumber,&ct);
      uu++;
    }
    if (istop) {
      //      cout << "This is a complex ridge!" << endl;
      mi = faces[jface].apex.coords;
      return true;
    }
  }
  return false;
}


bool PolytopeSmall::isSame(edge e1, edge e2) {
  return (e1.origin==e2.origin && e1.end == e2.end); 
}

bool PolytopeSmall::isFlipped(edge e1, edge e2) {
  return (e1.origin==e2.end && e1.end == e2.origin); 
}

bool PolytopeSmall::isEdgeNew(edge e) {
  bool res = true;
  unsigned l=0;
  while (res && l<edges.size()){
    res = !isSame(e,edges[l]) && !isFlipped(e,edges[l]);
    l++;
  }
  return res;
}

bool PolytopeSmall::testSame(WVector V, WVector W){
  return areDep(V.coords,W.coords);
}

bool PolytopeSmall::isInList(WVector V, vector<WVector> li){
  unsigned n = li.size();
  bool res = false;
  unsigned k = 0;
  while (!res && k<n) {
    res = testSame(V,li[k]);
    k++;
  }
  return res;
}

int PolytopeSmall::findInList(gen V, vector<WVector> li){
  vector<int> w;
  WVector W = WVector(V,w);
  return findInList(W,li);
}

int PolytopeSmall::findInList(WVector V, vector<WVector> li){
  unsigned n = li.size();
  bool res = false;
  int k = 0;
  while (!res && k<n) {
    res = testSame(V,li[k]);
    k++;
  }
  if (res) {
    return (k-1);
  }
  else {
    return -1;
  }
}

bool PolytopeSmall::computeVertices(Prism &A) {

  gen te = SqNorm(A.apex.coords);
  int res;
  if (checkSign(te,res)) {
    cout << "Sign of sq nm of apex: " << res << endl;

    unsigned k1 = 0;
    unsigned l1 = A.sides[0].word.size();
    for (unsigned k=1; k<A.sides.size(); k++) {
      unsigned l = A.sides[k].word.size();
      if (l<l1) {
	l1 = l;
	k1 = k;
      }
    }
    unsigned k2;
    if (k1==0) {
      if (A.sides[A.sides.size()-1].word.size()<A.sides[1].word.size()) {
	k2 = A.sides.size()-1;
      }
      else {
	k2 = 1;
      }
    }
    else {
      k2 = (k1+1)%A.sides.size();
    }
    
    gen M1 = reflMat(A.sides[k1].coords,mult)*reflMat(A.sides[k2].coords,mult);
    matred(M1);
    int om1 = order(M1);
    cout << "Product of order " << om1 << endl;
    int bo = braidOrder(reflMat(A.sides[k1].coords,mult),reflMat(A.sides[k2].coords,mult));
    cout << "Braiding order " << bo << endl;

    vector<int> w0 = concat(A.sides[k1].word,A.sides[k2].word);
    vector<int> w = concat(A.sides[k1].word,A.sides[k2].word);

    bool isrefl = false;
    unsigned k=1;
    gen M = M1;
    while (k<om1 && !isrefl) {
      gen dt = discrAdjusted(M);
      isrefl = operator_equal(dt,zeronumber,&ct);
      if (isrefl) {
	int sg;
	if (!checkSign(SqNorm(A.apex.coords),sg)) {
	  cerr << "Problem checking sign" << endl;
	  exit(1);
	}
	else {
	  if (sg>0) {
	    gen X1 = BoxProd(A.cspine,A.apex.coords);
	    gen X2 = M*X1;
	    vectred(X2);
	    isrefl = areDep(X1,X2);
	  }
	  else {
	    if (sg<0) {
	      gen X1 = A.apex.coords;
	      gen X2 = M*X1;
	      vectred(X2);
	      isrefl = areDep(X1,X2);
	    }
	    else {
	      // don't know what to do...
	    }
	  }
	}
      }
      if (isrefl) {
	cout << "Found reflection, power " << k << endl;
      }
      else {
	M = M*M1;
	w = concat(w,w0);
	matred(M);
      }
      k++;
    }
    if (!isrefl) {
      cerr << "Did not find any reflection power..." << endl;
      //      exit(1);
    }

    A.apex.word = w;
    cout << "w apex=" << convertword(w) << endl;

    if (res<1) {
      cout << "Single vertex at top" << endl;
      A.topface.push_back(A.apex);
    }
    else {
      cout << "Top face is not a point.." << endl;

      for (unsigned k=0; k<A.sides.size(); k++) {
	gen yo = BoxProd(A.apex.coords,A.sides[k].coords);
	WVector Vk( yo, concat(A.apex.word,A.sides[k].word) );
	A.topface.push_back(Vk);
        cout << "Vertex for " << convertword(A.sides[k].word) << endl;
	if (checkSign(SqNorm(Vk.coords),res)){
	  if (res>0) {
	    cerr << "This should not happen!" << endl;
	    exit(1);
	  }
	}
	else {
	  cerr << "Problem checking sign!" << endl;
	  exit(1);	  
	}	
      }
    }
  }
  else {
    cout << "Problem checking sign... " << evali(te) << endl;
  }

  cout << "Constructing bottom face..." << endl;
  for (unsigned k=0; k<A.sides.size(); k++) {
    gen yo = BoxProd(A.mirror.coords,A.sides[k].coords);
    if (checkSign(SqNorm(yo),res)){
      if (res>0) {
	cout << "#" << k << "(" << convertword(A.sides[k].word) << "): bottom and mid vertex." << endl;

	gen M1 = reflMat(A.mirror.coords,mult)*reflMat(A.sides[k].coords,mult);
	matred(M1);
	int om1 = order(M1);
	cout << "Product of order " << om1 << endl;
	int bo = braidOrder(reflMat(A.mirror.coords,mult),reflMat(A.sides[k].coords,mult));
	cout << "Braiding order " << bo << endl;

	vector<int> w0 = concat(A.mirror.word,A.sides[k].word);
	vector<int> wk = concat(A.mirror.word,A.sides[k].word);

	bool isrefl = false;
	unsigned kk=1;
	gen M = M1;
	while (kk<om1 && !isrefl) {
	  gen dt = discrAdjusted(M);
	  isrefl = operator_equal(dt,zeronumber,&ct);

	  if (isrefl) {
	    gen X1 = BoxProd(A.sides[k].coords,yo);
	    gen X2 = M*X1;
	    vectred(X2);
	    isrefl = areDep(X1,X2);
	  }
	  
	  // also want to check if vertex is fixed.
	  if (isrefl) {
	    cout << "Found reflection, power " << kk << endl;
	  }
	  M = M*M1;
	  wk = concat(wk,w0);
	  matred(M);
	  kk++;
	}
	if (!isrefl) {
	  cout << "Did not find any reflection power..." << endl;
	  cerr << "Did not find any reflection power..." << endl;
	  //	  exit(1);
	}

	WVector nm(yo,wk);

	gen yok1 = BoxProd(A.mirror.coords,yo);
	gen yok2 = BoxProd(A.sides[k].coords,yo);
	WVector Wk1(yok1,concat(A.mirror.word,wk));
	WVector Wk2(yok2,concat(A.sides[k].word,wk));
	if (checkSign(SqNorm(Wk1.coords),res)){
	  if (res>-1) {
	    cerr << "This should not happen!" << endl;
	    exit(1);
	  }
	}
	else {
	  cerr << "Problem checking sign!" << endl;
	  exit(1);	  
	}	
	if (checkSign(SqNorm(Wk2.coords),res)){
	  if (res>-1) {
	    cerr << "This should not happen!" << endl;
	    exit(1);
	  }
	}
	else {
	  cerr << "Problem checking sign!" << endl;
	  exit(1);	  
	}	
	A.bottomface.push_back(Wk1);
	A.midvertices.push_back(Wk2);
	A.midvertindices.push_back(k);
      }
      else {
	cout << "#" << k << "(" << convertword(A.sides[k].word) << "): Single bottom vertex." << endl;
	WVector Vk( yo, concat(A.mirror.word,A.sides[k].word) );
	A.bottomface.push_back(Vk);
      }
    }
    else {
      cerr << "Problem checking sign!" << endl;
      exit(1);	  
    }	
  }
  

}

bool PolytopeSmall::isOnMirror(gen X, Prism A) {
  gen te = Inn(X,A.mirror.coords);
  return operator_equal(te,zeronumber,&ct);  
}

bool PolytopeSmall::isOnTopFace(gen X, Prism A) {
  gen te = Inn(X,A.apex.coords);
  return operator_equal(te,zeronumber,&ct);  
}

bool PolytopeSmall::isOnPrism(gen X, Prism A){
  gen te = Inn(X,A.apex.coords)*Inn(A.apex.coords,A.apexProjection)*Inn(A.apexProjection,X);
  red(te);
  return isReal(te);
}

gen PolytopeSmall::reflectAcrossSpine(gen X, Prism A){
  vecteur co;
  co.push_back(Inn(A.apex.coords,A.apex.coords));
  co.push_back(Inn(A.apex.coords,A.apexProjection));
  co.push_back(Inn(A.apexProjection,A.apex.coords));
  co.push_back(Inn(A.apexProjection,A.apexProjection));
  gen M = createMatrix(2,2,co);

  vecteur cov;
  cov.push_back(Inn(X,A.apex.coords));
  cov.push_back(Inn(X,A.apexProjection));
  gen b = gen(cov);

  gen de = det(M);
  red(de);
  if (operator_equal(de,zeronumber,&ct)) {
    cerr << "Problem reflecting across spine" << endl;
    exit(1);
  }
  else {
    gen dei = inverse(de);
    vecteur coeffs;
    coeffs.push_back(dei*(M[1][1]*b[0]-M[0][1]*b[1]));
    coeffs.push_back(dei*(-M[1][0]*b[0]+M[0][0]*b[1]));
    gen Y = myconj(coeffs[0])*A.apex.coords + myconj(coeffs[1])*A.apexProjection;
    vectred(Y);
    return Y;
  }
}

vector<WVector> PolytopeSmall::findMirrors(gen V) {
  vector<WVector> res;
  for (unsigned k=0; k<faces.size(); k++) {
    if (isOnMirror(V,faces[k])) {
      if (findInList(faces[k].mirror,res)<0) {
	res.push_back(faces[k].mirror);
      }
    }
    if (isOnTopFace(V,faces[k])) {
      if (findInList(faces[k].apex,res)<0) {
	res.push_back(faces[k].apex);
      }
    }
  }
  return res;
}

vector<WVector> PolytopeSmall::findMirrors(WVector v) {
  return findMirrors(v.coords);
}

vector<int> PolytopeSmall::findFaces(gen V) {
  vector<int> res;
  gen te = SqNorm(V);
  if (operator_equal(te,zeronumber,&ct)) {
    res.push_back(-1);
  }
  for (unsigned k=0; k<faces.size(); k++) {
    if (isOnPrism(V,faces[k])) {
      bool fd = false;
      unsigned u=0;
      while (!fd && u<faces[k].topface.size()) {
	fd = areDep(V,faces[k].topface[u].coords);
	u++;
      }
      if (!fd) {
	u=0;
	while (!fd && u<faces[k].midvertices.size()) {
	  fd = areDep(V,faces[k].midvertices[u].coords);
	  u++;
	}
	if (!fd) {
	  u=0;
	  while (!fd && u<faces[k].bottomface.size()) {
	    fd = areDep(V,faces[k].bottomface[u].coords);
	    u++;
	  }
	  if (fd) {
	    res.push_back(k);
	  }
	}
	else {
	  res.push_back(k);
	}
      }
      else {
	res.push_back(k);
      }
    }
  }
  return res;
}

// beware this finds more faces than "findFaces", in general
vector<int> PolytopeSmall::findBisectors(gen V) {
  vector<int> res;
  gen te = SqNorm(V);
  if (operator_equal(te,zeronumber,&ct)) {
    res.push_back(-1);
  }
  for (unsigned k=0; k<faces.size(); k++) {
    if (isOnPrism(V,faces[k])) {
      res.push_back(k);
    }
  }
  return res;
}


vector<int> PolytopeSmall::findFaces(WVector v) {
  return findFaces(v.coords);
}

vector<int> PolytopeSmall::findBisectors(WVector v) {
  return findBisectors(v.coords);
}

void PolytopeSmall::useRelations(vector<int> &w) {
  unsigned n = w.size();
  bool reduced = false;
  while (n>0 && !reduced) {
    unsigned ns = w.size();
    for (unsigned k=0; k<relations.size(); k++)  {
      vector<int> rel = relations[k];
      if (2*w.size()>rel.size()) {
	simplifyword(w,rel);
      }
    }
    reduced = ns==w.size();
  }
}

void PolytopeSmall::useRelations(Prism &A) {
  useRelations(A.mirror.word);
  for (unsigned k=0; k<A.sides.size(); k++) {
    useRelations(A.sides[k].word);
  }
  for (unsigned k=0; k<A.bottomface.size(); k++) {
    useRelations(A.bottomface[k].word); 
 }
  for (unsigned k=0; k<A.topface.size(); k++) {
    useRelations(A.topface[k].word);
  }
  for (unsigned k=0; k<A.midvertices.size(); k++) {
    useRelations(A.midvertices[k].word);
  }
}

void PolytopeSmall::evalWord(vector<int> w, gen &res) {
  if (w.size()>0) {
    int k=w.size(); 
    while (k>0) {
      k--;
      int l = w[k];
      if (l==1) {
	res = R1*res;
      }
      else {
	if (l==2) {
	  res = R2*res;
	}
	else {
	  if (l==3) {
	    res = R3*res;
	  }
	  else {
	    if (l==-1) {
	      res = R1i*res;
	    }
	    else {
	      if (l==-2) {
		res = R2i*res;
	      }
	      else {
		if (l==-3) {
		  res = R3i*res;
		}
		else {
		  cerr << "Wrong index in word?" << endl;
		  exit(1);
		}
	      }
	    }
	  }
	}
      }
      vectred(res);
    }
  }
}

gen PolytopeSmall::evalWord(vector<int> w) {
  gen res = Id;
  unsigned k=0; 
  while (k<w.size()) {
    int l = w[k];
    if (l==1) {
      res = res*R1;
    }
    else {
      if (l==2) {
	res = res*R2;
      }
      else {
	if (l==3) {
	  res = res*R3;
	}
	else {
	  if (l==-1) {
	    res = res*R1i;
	  }
	  else {
	    if (l==-2) {
	      res = res*R2i;
	    }
	    else {
	      if (l==-3) {
		res = res*R3i;
	      }
	      else {
		cerr << "Wrong index in word?" << endl;
		exit(1);
	      }
	    }
	  }
	}
      }
    }
    matred(res);
    k++;
  }
  return res;
}

void PolytopeSmall::createBraidRelation(vector<int> w1, vector<int> w2) {
  gen A = evalWord(w1);
  gen B = evalWord(w2);
  //  cout << "Creating braid relation for " << w1 << " and " << w2 << endl;

  int bo = braidOrder(A,B);
  int bo2 = bo/2;

  cout << convertword(w1) << " and " << convertword(w2) << "  braid with order " << bo << endl;

  vector<int> w1i = invertword(w1);
  vector<int> w2i = invertword(w2);
  
  vector<int> r1;
  for (unsigned k=0; k<bo2; k++) {
    r1 = concat(r1,w1);
    r1 = concat(r1,w2);
  }
  if (bo%2!=0) {
    r1 = concat(r1,w1);
    r1 = concat(r1,w2i);
  }
  for (unsigned k=0; k<bo2; k++) {
    r1 = concat(r1,w1i);
    r1 = concat(r1,w2i);
  }

  vector<int> r2;
  for (unsigned k=0; k<bo2; k++) {
    r2 = concat(r2,w2);
    r2 = concat(r2,w1);
  }
  if (bo%2!=0) {
    r2 = concat(r2,w2);
    r2 = concat(r2,w1i);

  }
  for (unsigned k=0; k<bo2; k++) {
    r2 = concat(r2,w2i);
    r2 = concat(r2,w1i);
  }

  //  cout << "Created braid relation: " << r1 << endl;
  //  cout << "Created braid relation: " << r2 << endl;

  relations.push_back(r1);
  relations.push_back(r2);
  relations.push_back(shiftIndicesUp(r1));
  relations.push_back(shiftIndicesUp(r2));
  relations.push_back(shiftIndicesDown(r1));
  relations.push_back(shiftIndicesDown(r2));  

}

bool PolytopeSmall::isVertexSafe(WVector V, Prism A) {
  for (unsigned j=0; j<A.topface.size(); j++) {
    if (testSame(V,A.topface[j])) {
      return true;
    }
  }
  for (unsigned j=0; j<A.midvertices.size(); j++) {
    if (testSame(V,A.midvertices[j])) {
      return true;
    }
  }
  for (unsigned j=0; j<A.bottomface.size(); j++) {
    if (testSame(V,A.bottomface[j])) {
      return true;
    }
  }
  return false;  
}

void PolytopeSmall::checkEdge(unsigned j) {
  
  cout << "Studying edge #" << j << ", whose endpoints are vertices of: " << convertword(edges[j].faceindices) << "... ";

  gen X = vertices[edges[j].origin].coords;
  gen Y = vertices[edges[j].end].coords;
  /*  cout << "Origin is on faces: " << vertices[edges[j].origin].faceindices << endl;
  cout << "Origin is on bisectors: " << findBisectors(X) << endl;
  cout << "Endpt is on faces: " << vertices[edges[j].end].faceindices << endl;
  cout << "Endpt is on bisectors: " << findBisectors(Y) << endl;*/

  gen te = Inn(X,Y);
  Y = -te*Y;
  vectred(Y);
  gen Z = (1-t1)*X+t1*Y;
  vectred(Z);
  vector<int> inds;
  for (unsigned k=0; k<faces.size(); k++) {
    if (find(edges[j].faceindices.begin(),edges[j].faceindices.end(),k)==edges[j].faceindices.end()) {
      inds.push_back(k);
    }
  }

  for (unsigned k=0; k<inds.size(); k++) {
    gen eq = simplify(Inn(Z,faces[inds[k]].X0)*Inn(faces[inds[k]].X0,Z) - Inn(Z,faces[inds[k]].X1)*Inn(faces[inds[k]].X1,Z),&ct);
    red(eq);
    eq = convertToReal(eq);
    //    cout << "Equation #" << inds[k] << ": " << eq << endl;
    if (operator_equal(eq,zeronumber,&ct)) {
      cerr << "Equation is 0, this should not happen..." << endl;
      exit(1);
    }

    gen eq0 = _coeff(makesequence(eq,t1,0),&ct);
    gen eq1 = simplify(_subst(makesequence(eq,t1,onenumber),&ct),&ct);
    simplify(eq1,&ct);
    red(eq1);

    //    cout << "Values at endpoints: " << evali(eq0) << " ,  " << evali(eq1) << endl;

    int sgn0, sgn1;
    if (!checkSign(eq0,sgn0)){
      cerr << "Problem checking sign of equation at origin, in edge check" << endl;
      exit(1);      
    }
    if (!checkSign(eq1,sgn1)){
      cerr << "Problem checking sign of equation at endpt, in edge check" << endl;
      exit(1);      
    }
    if (sgn0>0 || sgn1>0) {
      cerr << "Edge is on WRONG side! (found this at an endpoint)" << endl;
      exit(1);      
    }

    if (sgn0==0){
      //      cout << "t=0 is a root" << endl;
      if (sgn1==0){
	cerr << "Both t=0 and t=1 are roots, this should mean edge is in the bisector!" << endl;
	exit(1);
      }
      else {
	gen a = _coeff(makesequence(eq,t1,2),&ct);
	if (operator_equal(a,zeronumber,&ct)) {
	  //	  cout << "Only t=0 gives a root" << endl;
	  //	  cout << "Edge is on negative side" << endl;
	  //	  return true;
	}
	else {
	  // check if te=-b/a is in [0,1]
	  gen te = -_coeff(makesequence(eq,t1,1),&ct)*inverse(a);
	  red(te);
	  //	  cout << "te=" << te << endl;
	  //	  cout << "Need to check if " << evali(te) << " is in [0,1]..." << endl;
	  int sc;
	  if (!checkSign(te,sc)) {
	    cerr << "Problem checking sign of t-value in edge parameter" << endl;
	    exit(1);
	  }
	  else {
	    if (sc>0) {
	      if (!checkSign(te-1,sc)) {
		cerr << "Problem checking whether t<=1 in edge parameter" << endl;
		exit(1);
	      }
	      else {
		if (sc<0) {
		  cerr << "The 1-skeleton is not embedded!" << endl;
		  exit(1);
		}
		/*	else {
		  return true;
		  }*/
	      }
	    }
	    else {
	      //  cout << "Only t=0 gives a root in [0,1]" << endl;	      
	      //  cout << "Edge is on negative side" << endl;
	      //	      return true;
	    }
	  }
	}	
      }
    }
    else {
      // sgn0 != 0
      if (sgn1==0){
	//	cout << "t=1 is a root" << endl;
	gen a = _coeff(makesequence(eq,t1,2),&ct);
	if (operator_equal(a,zeronumber,&ct)) {
	  //	  cout << "Only t=1 gives a root" << endl;
	  //	  return true;
	}
	else {
	  // check if te=-1-b/a is in [0,1]
	  gen te = -1-_coeff(makesequence(eq,t1,1),&ct)*inverse(a);
	  red(te);
	  //	  cout << "Need to check if " << evali(te) << " is in [0,1]..." << endl;
	  int sc;
	  if (!checkSign(te,sc)) {
	    cerr << "Problem checking sign of t-value in edge parameter" << endl;
	    exit(1);
	  }
	  else {
	    if (sc>0) {
	      if (!checkSign(te-1,sc)) {
		cerr << "Problem checking whether t<=1 in edge parameter" << endl;
		exit(1);
	      }
	      else {
		if (sc<0) {
		  cerr << "The 1-skeleton is not embedded!" << endl;
		  exit(1);
		}
		else {
		  //		  cout << "Only t=1 gives a root in [0,1]" << endl;	      
		  //		  cout << "Edge is on negative side" << endl;
		  //		  return true;
		}
	      }
	    }
	    else {
	      //	      cout << "Only t=1 gives a root in [0,1]" << endl;	      
	      //	      cout << "Edge is on negative side" << endl;
	      //	      return true;
	    }
	  }
	}	
      }
      else {
	// no endpoint is on bisector, need to work harder?
	// no, will simply evaluate value at maximum?  //2at+b -> -b/2a
	gen a = _coeff(makesequence(eq,t1,2),&ct);
	int sga;
	if (!checkSign(a,sga)) {
	  cerr << "Problem checking sign of a, in edge check" << endl;
	  exit(1);      
	}
	if (sga==0) {
	  //	  cout << "Linear equation, enough to check endpoints!" << endl;	      
	  //	  cout << "Edge is on negative side" << endl;
	  //	  return true;
	}
	else {
	  if (sga>0) {
	    //	    cout << "Concavity is up, checking endpoints is enough..." << endl;	    
	    //	    cout << "Edge is on negative side" << endl;
	    //	    return true;
	  }
	  else {
	    gen tm = -_coeff(makesequence(eq,t1,1),&ct)*inverse(2*a);
	    red(tm);

	    //	    cout << "t-max: " << evali(tm) << endl;

	    // check if max is in [0,1], and if so, check that the max value is negative
	    int sg;
	    if (!checkSign(tm,sg)) {
	      cerr << "Problem checking sign of t-value of max of parabola, in edge check" << endl;
	      exit(1);      	      
	    }
	    else {
	      if (sg>0) {
		// tmax is > 0
		if (!checkSign(tm-1,sg)) {
		  cerr << "Problem checking t-value of max of parabola, in edge check" << endl;
		  exit(1);      	      
		}
		else {
		  if (sg<0) {
		    // t-max is between 0 and 1...
		    gen yo = _subst(makesequence(eq,t1,tm),&ct);
		    if (!checkSign(yo,sg)) {
		      cerr << "Problem checking sign of max value, in edge check" << endl;
		      exit(1);      	      		      
		    }
		    else {
		      if (sg<=0) {
			//		cout << "Max of parabola occurs in [0,1], but value is negative." << endl;	    
			//		cout << "Edge is on negative side" << endl;
			//	    return true;
		      }
		      else {
			cerr << "Edge is on WRONG side! (found this at an endpoint)" << endl;
			exit(1);      
		      }
		    }
		  }
		  else {
		    // tmax is >1
		    //		    cout << "t-max is > 1" << endl;	    
		    //		    cout << "Edge is on negative side" << endl;
		  }
		}
	      }
	      else {
		// tmax is < 0
		//		cout << "t-max is < 0" << endl;	    
		//		cout << "Edge is on negative side" << endl;
	      }
	    }	    
	  }	  
	}
      }
    }
  }
  cout << " check!" << endl;
}

void PolytopeSmall::checkRidge(int j, int k) {

  cout << endl;
  cout << "Checking ridge on faces #" << j << " and " << k << endl;

  bool genparam = true;
  gen X0 = BoxProd(faces[j].cspine,faces[k].cspine);
  gen V0, W0, Z0; // Orthogonal (but not orthonormal) basis, with V0, W0 spanning common slice
  if (operator_equal(X0,zerovector,&ct)) {
    // cospinal
    genparam = false;
    gen Xte = BoxProd(faces[j].mirror.coords,faces[k].mirror.coords);
    if (operator_equal(Xte,zerovector,&ct)) {
      cout << "Same mirror, this gives the common slice" << endl;
      V0 = faces[j].apexProjection;
      //      W0 = Proj(faces[j].bottomface[0].coords,V0);	
      //    Z0 = BoxProd(V0,W0); // not clear if we need this
    }
    else {
      Xte = BoxProd(faces[j].mirror.coords,faces[k].apex.coords);
      if (operator_equal(Xte,zerovector,&ct)) {
	cout << "Mirror of face#" << j << " is same as top face of #" << k << endl;
	cout << "     this gives the common slice" << endl;	
	V0 = faces[j].apexProjection;
	//	W0 = Proj(faces[j].bottomface[0].coords,V0);
	//	Z0 = BoxProd(V0,W0); // not clear if we need this
      }
      else {
	Xte = BoxProd(faces[k].mirror.coords,faces[j].apex.coords);
	if (operator_equal(Xte,zerovector,&ct)) {
	  cout << "Mirror of face#" << k << " is same as top face of #" << j << endl;
	  cout << "     this gives the common slice" << endl;	
	  V0 = faces[k].apexProjection;
	  //  W0 = Proj(faces[k].bottomface[0].coords,V0);
	  //  Z0 = BoxProd(V0,W0); // not clear if we need this
	}
	else {
	  cerr << "There is probably no common slice, will need to check this.." << endl;	  
	  //	  exit(1);
	}
      }
    }
  }
  else {
    // complex spines intersect in projective space, need to check if they intersect on the real spines
    if (isOnPrism(X0,faces[j])) {
      cout << "Extended real spine of face #" << j << " intersects extor #" << k << endl;
      if (isOnPrism(X0,faces[k])) {
	cout << "Extended real spine of face #" << k << " intersects extor #" << j << endl;
	cout << "Real spines intersect... "; //either cotranchal, or linear?
	int sx;
	if (!checkSign(SqNorm(X0),sx)) {
	  cerr << "Trouble computing sign of square norm (in checkRidge)" << endl;
	  exit(1);
	}
	else {
	  if (sx>0) {
	    cout << "Cotranchal pair" << endl;

	    gen Y1 = faces[j].apexProjection;
	    Y1 = Inn(X0,Y1)*Y1;
	    red(Y1);
	    Y1 = Proj(Y1,X0);
	    gen Y2 = faces[k].apexProjection;
	    Y2 = Inn(X0,Y2)*Y2;
	    red(Y2);
	    Y2 = Proj(Y2,X0);
	    gen tee = Inn(Y1,faces[j].X0)*Inn(faces[j].X0,Y1)-Inn(Y1,faces[j].X1)*Inn(faces[j].X1,Y1);
	    red(tee);
	    tee = Inn(Y2,faces[k].X0)*Inn(faces[k].X0,Y2)-Inn(Y2,faces[k].X1)*Inn(faces[k].X1,Y2);
	    red(tee);
	    gen tt = (Inn(Y1,Y2)+Inn(Y2,Y1))/2;
	    gen te = SqNorm(Y1)*SqNorm(Y2) - tt*tt;
	    red(te);
	    te = convertToReal(te);
	    int sgn;
	    if (!checkSign(te,sgn)) {
	      cerr << "Trouble checking sign (is intersection is slice only)" << endl;
	      exit(1);
	    }
	    else {
	      if (sgn<0) {
		cerr << "Intersection is not just the common slice!?" << endl;
		exit(1);
	      }
	      else {
		cout << "intersection is slice only!" << endl;
		genparam = false;
	      }
	    }
	      
	    V0 = X0;
	    if (areDep(X0,faces[j].apex.coords)) {
	      cout << "Common slice is top face of face #" << j << endl;
	      W0 = Proj(faces[j].topface[0].coords,V0);
	    }
	    else {
	      if (areDep(X0,faces[j].mirror.coords)) {
		cout << "Common slice is mirror of face #" << j << endl;
		W0 = Proj(faces[j].bottomface[0].coords,V0);
	      }
	      else {
		cerr << "Unidentified common slice, this is odd.." << endl;
		exit(1);
	      }
	    }
	    Z0 = BoxProd(V0,W0); // not clear if we need this
	    if (areDep(X0,faces[k].apex.coords)) {
	      cout << "Common slice is top face of face #" << k << endl;
	    }
	    else {
	      if (areDep(X0,faces[k].mirror.coords)) {
		cout << "Common slice is mirror of face #" << k << endl;
	      }
	      else {
		cerr << "Unidentified common slice, this is odd.." << endl;
		exit(1);
	      }
	    }
	  }
	  else {
	    if (sx==0) {
	      cerr << "Strange, real spines intersect at infinity..." << endl;
	      exit(1);
	    }
	    else {
	      cerr << "Real spines intersect inside, linear intersection (not implemented yet)" << endl;
	      exit(1);
	    }
	  }
	}
      }	
      else {
	cout << "Real spines don't intersect" << endl;
	genparam = true;
      }
    }
    else {
      cout << "Real spines don't intersect" << endl;
      genparam = true;
    }
  }

  if (genparam) {
    gen vj, wj, vk, wk, vl, wl;

    wj = faces[j].apexProjection;
    vj = faces[j].apex.coords;
    vj = Inn(wj,vj)*vj;
    vectred(vj);
    vj = Proj(vj,wj);

    wk = faces[k].apexProjection;
    vk = faces[k].apex.coords;
    vk = Inn(wk,vk)*vk;
    vectred(vk);
    vk = Proj(vk,wk);

    // need vl and wl, for third bisector in Giraud's theorem!

    cout << "Will construct third bisector (Giraud)" << endl;

    bool fd = false;
    int u1=-1;
    while (!fd && u1!=faces[j].sides.size()-1) {
      u1++;
      fd = areDep(faces[j].sides[u1].coords,faces[k].mirror.coords);
    }
    if (!fd) {
      cerr << "Trouble findind third bisector (mirror)" << endl;
      exit(1);
    }

    int u2=-1;
    int v;
    fd = false;
    while (!fd && u2!=faces[j].sides.size()-1) {
      u2++;
      v=-1;
      while (!fd && v!=faces[k].sides.size()-1) {
	v++;
	fd = areDep(faces[j].sides[u2].coords,faces[k].sides[v].coords);
      }
    }
    if (!fd) {
      cerr << "Trouble findind third bisector (side)" << endl;
      exit(1);
    }

    // mirror and apex of third bisector
    gen mi = faces[j].sides[u2].coords;
    gen ap = BoxProd(faces[j].sides[u1].coords,faces[j].mirror.coords);
    gen app = Proj(ap,mi);
    
    wl = app;
    vl = Inn(app,ap)*ap;
    vectred(vl);
    vl = Proj(vl,wl);

    gen t1min, t1max, t2min, t2max, t3min, t3max;
    // will need to start finding the vertices!

    computeVertices(faces[j]);
    //    computeVertices(faces[k]);
    vector<WVector> list;
    for (unsigned uu=0; uu<faces[j].bottomface.size(); uu++) {
      WVector X = faces[j].bottomface[uu];
      if (isVertexSafe(X,faces[k])) {
	list.push_back(X);
      }
    }

    for (unsigned z=0; z<list.size(); z++) {
      //      cout << "Testing vertex #" << ridges[i].verts[z] << endl;
      gen t1v = -Inn(vj,list[z].coords)*inverse(Inn(wj,list[z].coords));
      gen t2v = -Inn(vk,list[z].coords)*inverse(Inn(wk,list[z].coords));
      gen t3v = -Inn(vl,list[z].coords)*inverse(Inn(wl,list[z].coords));
      red(t1v);
      red(t2v);
      red(t3v);
      t1v = convertToReal(t1v);
      t2v = convertToReal(t2v);
      t3v = convertToReal(t3v);
      if (z==0) {
	t1min = t1v;
	t1max = t1v;
	t2min = t2v;
	t2max = t2v;
	t3min = t3v;
	t3max = t3v;
      }
      else {
	int sg;
	if (!checkSign(t1v-t1min,sg)) {
	  cerr << "Problem checking sign (t1min)" << endl;
	}
	else {
	  if (sg<0) {
	    t1min = t1v;
	  }
	}
	if (!checkSign(t1v-t1max,sg)) {
	  cerr << "Problem checking sign (t1max)" << endl;
	}
	else {
	  if (sg>0) {
	    t1max = t1v;
	  }
	}

	if (!checkSign(t2v-t2min,sg)) {
	  cerr << "Problem checking sign (t2min)" << endl;
	}
	else {
	  if (sg<0) {
	    t2min = t2v;
	  }
	}
	if (!checkSign(t2v-t2max,sg)) {
	  cerr << "Problem checking sign (t2max)" << endl;
	}
	else {
	  if (sg>0) {
	    t2max = t2v;
	  }
	}

	if (!checkSign(t3v-t3min,sg)) {
	  cerr << "Problem checking sign (t3min)" << endl;
	}
	else {
	  if (sg<0) {
	    t3min = t3v;
	  }
	}
	if (!checkSign(t3v-t3max,sg)) {
	  cerr << "Problem checking sign (t3max)" << endl;
	}
	else {
	  if (sg>0) {
	    t3max = t3v;
	  }
	}
      }
    }
    /*    gen t1mini = evali(t1min);
    gen t1maxi = evali(t1max);
    gen t2mini = evali(t2min);
    gen t2maxi = evali(t2max);
    gen t3mini = evali(t3min);
    gen t3maxi = evali(t3max);
    // verbose?
    cout << "t1 range: " << t1mini << " , " << t1maxi << endl;
    cout << "t2 range: " << t2mini << " , " << t2maxi << endl;
    cout << "t3 range: " << t3mini << " , " << t3maxi << endl;*/
    
    // take (vj+t1*wj)x(vk+t2*wk)
    gen v0, v1, v2, v3; 
    v0 = BoxProd(vj,vk);
    v1 = BoxProd(wj,vk);
    v2 = BoxProd(vj,wk);
    v3 = BoxProd(wj,wk);
    gen vt = v0 + t1*v1 + t2*v2 + t1*t2*v3;
    //    gen te = SqNorm(vt);
    //    cout << "Eq bdy: " << te << endl;

    gen t3n = -Inn(vt,vl);
    gen t3d = Inn(vt,wl);

    for (unsigned l=0; l<faces.size(); l++) {
      gen eq = Inn(vt,faces[l].X0)*Inn(faces[l].X0,vt) - Inn(vt,faces[l].X1)*Inn(faces[l].X1,vt); 
      red(eq);
      eq = convertToReal(eq);
      
      if (operator_equal(eq,zeronumber,&ct)) {
	cout << "Eq #" << l << " is 0..." << endl;
      }
      else {
	//	cout << "Eq #" << l << ": " << eq << endl; 
	gen eq1 = _derive(makesequence(eq,t1),&ct);
	gen eq2 = _derive(makesequence(eq,t2),&ct);
	
	vecteur eqs_rur;
	eqs_rur.push_back(eq1);
	eqs_rur.push_back(eq2);
	eqs_rur.push_back(realminpol);
	
	//	cout << "Starting computation of RUR..." << endl;

	gen gb1 = _gbasis(makesequence(eqs_rur,vars2d,change_subtype(_RUR_REVLEX,_INT_GROEBNER)),&ct);

	//	cout << "gb1=" << gb1 << endl;

	if (gb1[0]==-4) {

	  vecteur valpars;
	  if (!safeSift(gb1,valpars)) {
	    cerr << "Problem sifting RUR values... (crit pts)" << endl;
	    exit(1);
	  }
	  else {
	    cout << "For eq#" << l << ", found " << valpars.size() << " critical point(s).. " << endl; 
	    
	    for (unsigned u = 0; u<valpars.size(); u++) {
	      
	      cout << "Checking crit pt #" << u << "..  ";
	      
	      gen t1vi = _horner(makesequence(gb1[5],valpars[u],s__IDNT_e),&ct)/_horner(makesequence(gb1[3],valpars[u],s__IDNT_e),&ct);
              gen t2vi = _horner(makesequence(gb1[6],valpars[u],s__IDNT_e),&ct)/_horner(makesequence(gb1[3],valpars[u],s__IDNT_e),&ct);
	      // verbose??
	      //	      cout << t1vi << " , " << t2vi << endl;
	      	      

	      int stest;
	      if (!checkt1val(gb1,valpars[u],t1min,stest)) {
		cerr << "Trouble comparing with t1min" << endl;
		exit(1);
	      }
	      if (stest<0) {
		cout << "outside (t1<t1min)" << endl;
	      }
	      else {
		if (!checkt1val(gb1,valpars[u],t1max,stest)) {
		  cerr << "Trouble comparing with t1max" << endl;
		  exit(1);
		}
		if (stest>0) {
		  cout << "outside (t1>t1max)" << endl;
		}
		else {
		  if (!checkt2val(gb1,valpars[u],t2min,stest)) {
		    cerr << "Trouble comparing with t2min" << endl;
		    exit(1);
		  }
		  if (stest<0) {
		    cout << "outside (t2<t2min)" << endl;
		  }
		  else {
		    if (!checkt2val(gb1,valpars[u],t2max,stest)) {
		      cerr << "Trouble comparing with t2max" << endl;
		      exit(1);
		    }
		    if (stest>0) {
		      cout << "outside (t2>t2max)" << endl;
		    }
		    else {
		      if (!checkt3val(gb1,valpars[u],t3min,vl,wl,vt,stest)) {
			cerr << "Trouble comparing with t3min" << endl;
			exit(1);
		      }
		      if (stest<0) {
			cout << "outside (t3<t3min)" << endl;
		      }
		      else {
			if (!checkt3val(gb1,valpars[u],t3max,vl,wl,vt,stest)) {
			  cerr << "Trouble comparing with t3max" << endl;
			  exit(1);
			}
			if (stest>0) {
			  cout << "outside (t3>t3max)" << endl;
			}
			else {
			  if (!checkEqSign(gb1,valpars[u],eq,stest)) {
			    cerr << "Trouble checking sign of equation" << endl;
			    exit(1);
			  }
			  if (eq>0) {
			    cerr << "Equation is positive inside the G-polygon..." << endl;
			    cerr << "There seems to be an isolated component!!!!!!!" << endl;
			    exit(1);
			  }
			  else {
			    cout << "CRITICAL POINT IS INSIDE, but the equation is negative there!" << endl;
			  }
			}
		      }
		    }
		  }
		}
	      }
	    }	      
	  }
	}
	else {
	  cout << "Non isolated critical points? the system is not 0-dim" << endl;
	  // could either check if eqn for geodesics give (multiple) factors, or
	  //    use factorization over a number field, or factorization by hand... 
	  // For now, I choose the last option.

	  bool needmore = false;

	  gen co = _coeff(makesequence(eq,t2),&ct);

	  gen a = _coeff(makesequence(co[0],t1,2),&ct);
	  gen b = _coeff(makesequence(co[1],t1,2),&ct);
	  gen c = _coeff(makesequence(co[2],t1,2),&ct);

	  gen te1,te2;
	  te1 = _rem(makesequence(a*co[1]-b*co[0],realminpol,s__IDNT_e),&ct);
	  if (operator_equal(te1,zeronumber,&ct)) {
	    te2 = _rem(makesequence(a*co[2]-c*co[0],realminpol,s__IDNT_e),&ct);
	    if (operator_equal(te2,zeronumber,&ct)) {
	      gen discr = _rem(makesequence(b*b-4*a*c,realminpol,s__IDNT_e),&ct);
	      if (operator_equal(discr,zeronumber,&ct)) {
		cout << "There is a square factor of the form (t2-r)^2" << endl;
		gen r = -b*inverse(2*a);
		red(r);
		cout << "  r = " << r << endl;
		int sg;
		gen zz = r-t2min;
		red(zz);
		//		cout << "r-t2min=" << zz << endl;
		if (checkSign(zz,sg)){
		  if (sg>0) {
		    cout << "r>t2min" << endl;
		    zz = r-t2max;
		    red(zz);
		    //		    cout << "r-t2max=" << zz << endl;
		    if (checkSign(zz,sg)){
		      if (sg<0) {
			cout << "r<t2max" << endl;
			cout << "Horizontal line crosses the interior of the ridge!" << endl;
			exit(1);
		      }
		    }
		    else {
		      cerr << "Trouble comparing r with t2max" << endl;
		      exit(1);
		    }
		  }
		}
		else {
		  cerr << "Trouble comparing r with t2min" << endl;
		  exit(1);
		}

		// need to check sign of co[0], check it is positive in [t1min,t1max]!
		gen yo1 = _subst(makesequence(co[0],t1,t1min),&ct);
		red(yo1);
		//		cout << "yo1=" << yo1 << endl;
		if (checkSign(yo1,sg)) {
		  if (sg>0) {
		    cerr << "Equation seems to cut into the ridge, have you checked edges?" << endl;
		    exit(1);
		  }
		  else {
		    cout << "g(t1)==eq/(t2-r)^2 is <=0 at t1min" << endl;
		    gen yo2 = _subst(makesequence(co[0],t1,t1max),&ct);
		    red(yo2);
		    //		    cout << "yo2=" << yo2 << endl;
		    if (checkSign(yo2,sg)) {
		      if (sg>0) {
			cout << "g(t1)==eq/(t2-r)^2 is <=0 at t1max" << endl;
			cerr << "Equation seems to cut into the ridge, have you checked edges?" << endl;
			exit(1);
		      }
		      else {
			// co[0] negative at enpoints, if degree two then still something to do!
			cout << "co[0] is neg at endpoints, but need to check more" << endl;
			cout << "co[0]=" << co[0] << endl;
			gen de = _degree(makesequence(co[0],t1),&ct);
			cout << "degree: " << de << endl;
			if (operator_equal(de,twonumber,&ct)) {
			  gen aa = _coeff(makesequence(co[0],t1,2),&ct);
			  red(aa);
			  //			  cout << "aa=" << aa << endl;
			  if (checkSign(aa,sg)) {
			    if (sg<0) {
			      cout << "coeff of t1^2 is <0, will check tv = t1 of vertex of parabola" << endl;
			      gen tmax = _coeff(makesequence(co[0],t1,1),&ct)*inverse(2*_coeff(makesequence(co[0],t1,2),&ct));
			      red(tmax);
			      gen zzz = tmax-t1min;
			      red(zzz);
			      //			      cout << "tver-t1min=" << zz << endl;
			      if (checkSign(zzz,sg)) {
				if (sg>0) {
				  cout << "tv>t1min" << endl;
				  zzz = tmax-t1max;
				  red(zzz);
				  //				  cout << "tver-t1max=" << zz << endl;
				  if (checkSign(zzz,sg)) {
				    if (sg<0) {
				      cout << "tv<t1max" << endl;
				      gen val = _subst(makesequence(co[0],t1,tmax),&ct);
				      red(val);
				      cout << "val at vertex of parabola: " << val << endl;
				      if (checkSign(val,sg)) {
					if (sg>0) {
					  cout << "Value at tv is > 0!!!" << endl;
					  cout << "Polytope does not seem embedded, sorry!" << endl;
					  exit(1);
					}
					else {
					  cout << "Value at tv is <= 0, OK." << endl;
					}
				      }
				      else {
					cerr << "Trouble checking sign at vertex of parabola." << endl;
					exit(1);
				      }
				    }
				  }
				  else {
				    cerr << "Trouble comparing t-vertex with t1max." << endl;
				    exit(1);
				  }
				}
			      }
			      else {
				cerr << "Trouble comparing t-vertex with t1min." << endl;
				exit(1);
			      }
			    }
			  }
			  else {
			    cerr << "Trouble checking sign of coeff." << endl;
			    exit(1);
			  }
			}
		      }
		    }
		    else {
		      cerr << "Trouble checking sign of coeff at t1min" << endl;
		      exit(1);		  
		    }
		  }
		}
		else {
		  cerr << "Trouble checking sign of coeff at t1min" << endl;
		  exit(1);		  
		}
	      }
	      else {
		needmore = true;
	      }
	    }
	    else {
	      needmore = true;
	    }
	  }
	  else {
	    needmore = true;
	  }

	  if (needmore) {
	    // start over with other variable!

	    bool needevenmore = false;

	    gen co = _coeff(makesequence(eq,t1),&ct);

	    gen a = _coeff(makesequence(co[0],t2,2),&ct);
	    gen b = _coeff(makesequence(co[1],t2,2),&ct);
	    gen c = _coeff(makesequence(co[2],t2,2),&ct);

	    gen te1,te2;
	    te1 = _rem(makesequence(a*co[1]-b*co[0],realminpol,s__IDNT_e),&ct);
	    if (operator_equal(te1,zeronumber,&ct)) {
	      te2 = _rem(makesequence(a*co[2]-c*co[0],realminpol,s__IDNT_e),&ct);
	      if (operator_equal(te2,zeronumber,&ct)) {
		gen discr = _rem(makesequence(b*b-4*a*c,realminpol,s__IDNT_e),&ct);
		if (operator_equal(discr,zeronumber,&ct)) {
		  cout << "There is a square factor of the form (t1-r)^2" << endl;
		  gen r = -b*inverse(2*a);
		  red(r);
		  cout << "  r = " << r << endl;
		  int sg;
		  gen zzzz = r-t1min;
		  red(zzzz);
		  //		  cout << "r-t1min=" << zzzz << endl;
		  if (checkSign(zzzz,sg)){
		    if (sg>0) {
		      zzzz = r-t1max;
		      red(zzzz);
		      //		      cout << "r-t1max=" << zzzz << endl;
		      if (checkSign(zzzz,sg)){
			if (sg<0) {
			  cout << "Vertical line crosses the interior of the ridge!" << endl;
			  exit(1);
			}
		      }
		      else {
			cerr << "Trouble comparing r with t1max" << endl;
			exit(1);
		      }
		    }
		  }
		  else {
		    cerr << "Trouble comparing r with t1min" << endl;
		    exit(1);
		  }

		  // need to check sign of co[0], check it is positive in [t1min,t1max]!
		  gen yo1 = _subst(makesequence(co[0],t2,t2min),&ct);
		  red(yo1);
		  //	  cout << "yo1=" << yo1 << endl;
		  if (checkSign(yo1,sg)) {
		    if (sg>0) {
		      cerr << "Equation seems to cut into the ridge, have you checked edges?" << endl;
		      exit(1);
		    }
		    else {
		      gen yo2 = _subst(makesequence(co[0],t2,t2max),&ct);
		      red(yo2);
		      //     cout << "yo2=" << yo2 << endl;
		      if (checkSign(yo2,sg)) {
			if (sg>0) {
			  cerr << "Equation seems to cut into the ridge, have you checked edges?" << endl;
			  exit(1);
			}
			else {
			  // co[0] negative at enpoints, if degree two then still something to do!
			  //	  cout << "co[0]=" << co[0] << endl;
			  gen de = _degree(makesequence(co[0],t2),&ct);
			  //	  cout << "degree: " << de <<  endl;
			  if (operator_equal(de,twonumber,&ct)) {
			    gen aa = _coeff(makesequence(co[0],t2,2),&ct);
			    red(aa);
			    //    cout << "aa=" << aa << endl;
			    if (checkSign(aa,sg)) {
			      if (sg<0) {
				gen tmax = _coeff(makesequence(co[0],t2,1),&ct)*inverse(2*_coeff(makesequence(co[0],t2,2),&ct));
				red(tmax);
				//		cout << "t-vert=" << tmax << endl;
				gen zzzzz = tmax-t2min;
				red(zzzzz);
				//				cout << "tmax-t2min=" << zzzzz << endl;
				if (checkSign(tmax-t2min,sg)) {
				  if (sg>0) {
				    zzzzz = tmax-t2max;
				    red(zzzzz);
				    //	    cout << "tmax-t2max=" << zzzzz << endl;				    
				    if (checkSign(tmax-t2max,sg)) {
				      if (sg<0) {
					gen val = _subst(makesequence(co[0],t2,tmax),&ct);
					red(val);
					cout << "Val at vert: " << val << endl;
					if (checkSign(val,sg)) {
					  if (sg>0) {
					    cout << "Polytope does not seem embedded, sorry!" << endl;
					    exit(1);
					  }
					}
					else {
					  cerr << "Trouble checking sign at vertex of parabola." << endl;
					  exit(1);
					}
				      }
				    }
				    else {
				      cerr << "Trouble comparing t-vertex with t1max." << endl;
				      exit(1);
				    }
				  }
				}
				else {
				  cerr << "Trouble comparing t-vertex with t1min." << endl;
				  exit(1);
				}
			      }
			    }
			    else {
			      cerr << "Trouble checking sign of coeff." << endl;
			      exit(1);
			    }
			  }
			}
		      }
		      else {
			cerr << "Trouble checking sign of coeff at t1min" << endl;
			exit(1);		  
		      }
		    }
		  }
		  else {
		    cerr << "Trouble checking sign of coeff at t1min" << endl;
		    exit(1);		  
		  }
		}
		else {
		  needevenmore = true;
		}
	      }
	      else {
		needevenmore = true;
	      }
	    }
	    else {
	      needevenmore = true;
	    }

	    if (needevenmore) {
	      cerr << "Factors (t1-r)^2 or (t2-r)^2 did not explain not 0-dim" << endl;
	      cerr << "It may help to factor over number field, or use t3?" << endl;
	      exit(1);
	    }
	  }
	}
      }
    }
    
  }
  else {
    // use V0, W0, Z0
    cout << "For this ridge there is nothing to do, it is enough to check the 1-skeleton!" << endl;
  }
}

void PolytopeSmall::drawRidgePic(unsigned i) {
  int j =  ridges[i].faceindices[0];
  int k =  ridges[i].faceindices[1];
  cout << "Trying to draw pic of ridge on faces #" << j << " and " << k << endl;
  cout << " (I will only do it provided this ridge is a complex disk)" << endl;

  bool genparam;
  gen X0 = BoxProd(faces[j].cspine,faces[k].cspine);
  gen V0, W0, Z0; // Orthogonal (but not orthonormal) basis, with V0, W0 spanning common slice
  if (operator_equal(X0,zerovector,&ct)) {
    // cospinal
    genparam = false;
    gen Xte = BoxProd(faces[j].mirror.coords,faces[k].mirror.coords);
    if (operator_equal(Xte,zerovector,&ct)) {
      cout << "Same mirror, this gives the common slice" << endl;
      V0 = faces[j].apexProjection;
      W0 = Proj(faces[j].bottomface[0].coords,V0);	
      Z0 = BoxProd(V0,W0); // not clear if we need this
    }
    else {
      Xte = BoxProd(faces[j].mirror.coords,faces[k].apex.coords);
      if (operator_equal(Xte,zerovector,&ct)) {
	cout << "Mirror of face#" << j << " is same as top face of #" << k << endl;
	cout << "     this gives the common slice" << endl;	
	V0 = faces[j].apexProjection;
	W0 = Proj(faces[j].bottomface[0].coords,V0);
	Z0 = BoxProd(V0,W0); // not clear if we need this
      }
      else {
	Xte = BoxProd(faces[k].mirror.coords,faces[j].apex.coords);
	if (operator_equal(Xte,zerovector,&ct)) {
	  cout << "Mirror of face#" << k << " is same as top face of #" << j << endl;
	  cout << "     this gives the common slice" << endl;	
	  V0 = faces[k].apexProjection;
	  W0 = Proj(faces[k].bottomface[0].coords,V0);
	  Z0 = BoxProd(V0,W0); // not clear if we need this
	}
	else {
	  cerr << "There is probably no common slice, will need to check this.." << endl;
	  exit(1);
	}
      }
    }
  }
  else {
    // complex spines intersect in projective space, need to check if they intersect on the real spines

    if (isOnPrism(X0,faces[j])) {
      cout << "Extended real spine of face #" << j << " intersects extor #" << k << endl;
      if (isOnPrism(X0,faces[k])) {
	cout << "Extended real spine of face #" << k << " intersects extor #" << j << endl;
	cout << "Real spines intersect... "; //either cotranchal, or linear?
	int sx;
	if (!checkSign(SqNorm(X0),sx)) {
	  cerr << "Trouble computing sign of square norm (in checkRidge)" << endl;
	  exit(1);
	}
	else {
	  if (sx>0) {
	    cout << "Cotranchal pair" << endl;

	    gen Y1 = faces[j].apexProjection;
	    Y1 = Inn(X0,Y1)*Y1;
	    red(Y1);
	    Y1 = Proj(Y1,X0);
	    gen Y2 = faces[k].apexProjection;
	    Y2 = Inn(X0,Y2)*Y2;
	    red(Y2);
	    Y2 = Proj(Y2,X0);
	    gen tee = Inn(Y1,faces[j].X0)*Inn(faces[j].X0,Y1)-Inn(Y1,faces[j].X1)*Inn(faces[j].X1,Y1);
	    red(tee);
	    tee = Inn(Y2,faces[k].X0)*Inn(faces[k].X0,Y2)-Inn(Y2,faces[k].X1)*Inn(faces[k].X1,Y2);
	    red(tee);
	    gen tt = (Inn(Y1,Y2)+Inn(Y2,Y1))/2;
	    gen te = SqNorm(Y1)*SqNorm(Y2) - tt*tt;
	    red(te);
	    te = convertToReal(te);
	    int sgn;
	    if (!checkSign(te,sgn)) {
	      cerr << "Trouble checking sign (is intersection is slice only)" << endl;
	      exit(1);
	    }
	    else {
	      if (sgn<0) {
		cerr << "Intersection is not just the common slice!?" << endl;
		exit(1);
	      }
	      else {
		cout << "intersection is slice only!" << endl;
		genparam = false;
	      }
	    }
	      
	    V0 = X0;
	    if (areDep(X0,faces[j].apex.coords)) {
	      cout << "Common slice is top face of face #" << j << endl;
	      W0 = Proj(faces[j].topface[0].coords,V0);
	    }
	    else {
	      if (areDep(X0,faces[j].mirror.coords)) {
		cout << "Common slice is mirror of face #" << j << endl;
		W0 = Proj(faces[j].bottomface[0].coords,V0);
	      }
	      else {
		cerr << "Unidentified common slice, this is odd.." << endl;
		exit(1);
	      }
	    }
	    Z0 = BoxProd(V0,W0); // not clear if we need this
	    if (areDep(X0,faces[k].apex.coords)) {
	      cout << "Common slice is top face of face #" << k << endl;
	    }
	    else {
	      if (areDep(X0,faces[k].mirror.coords)) {
		cout << "Common slice is mirror of face #" << k << endl;
	      }
	      else {
		cerr << "Unidentified common slice, this is odd.." << endl;
		exit(1);
	      }
	    }
	  }
	  else {
	    if (sx==0) {
	      cerr << "Strange, real spines intersect at infinity..." << endl;
	      exit(1);
	    }
	    else {
	      cerr << "Real spines intersect inside, linear intersection (not implemented yet)" << endl;
	      exit(1);
	    }
	  }
	}
      }	
      else {
	cerr << "Real spines don't intersect" << endl;
	genparam = true;
      }
    }
    else {
      cerr << "Real spines don't intersect" << endl;
      genparam = true;
    }
  }

  if (genparam) {
    cout << "This is a generic ridge, picture not implemented yet" << endl;
  }
  else {
    ofstream file;
    string jc = static_cast<ostringstream*>( &(ostringstream() << j) )->str();
    string kc = static_cast<ostringstream*>( &(ostringstream() << k) )->str();
    string filename = "pic-ridge"+tag+"_"+jc+"-"+kc+".asy";
    file.open(filename.c_str());
    file << "import graph;" << endl;
    file << "size(7.5cm,0);" << endl;
    file << "draw(circle((0,0),1),red);" << endl;

    vector<int> inds = findFaces(V0);
    gen V0n = evali(V0);
    gen W0n = evali(W0);
    gen Hn = evali(H);

    gen X1 = evali(vertices[ridges[i].verts[0]].coords);
    gen X2 = evali(vertices[ridges[i].verts[3]].coords);
    X1 = X1/sqrt(-re(conj(X1,&ct)*Hn*X1,&ct),&ct);
    X2 = X2/sqrt(-re(conj(X2,&ct)*Hn*X2,&ct),&ct);
    gen yo = conj(X2,&ct)*Hn*X1;
    X2 = (-yo/abs(yo,&ct))*X2;
    gen MP = X1+X2;
    gen mpz = (-conj(MP,&ct)*Hn*W0n)/(conj(MP,&ct)*Hn*V0n);

    W0n = (conj(mpz,&ct)/abs(mpz,&ct))*W0n;
    V0n = V0n/sqrt(-re(conj(V0n,&ct)*Hn*V0n,&ct),&ct);
    W0n = W0n/sqrt(re(conj(W0n,&ct)*Hn*W0n,&ct),&ct);

    gen mpz2 = (-conj(MP,&ct)*Hn*W0n)/(conj(MP,&ct)*Hn*V0n);
    //    cout << "This should be real: " << mpz2 << endl;
    //    file << "label(\"mp\",(" << re(mpz2,&ct) << ",0));" << endl;

    //    cout << "V0n = " << V0n << endl;
    //    cout << "W0n = " << W0n << endl;
    //   cout << "Hn = " << Hn << endl;

    for (int u=0; u<faces.size(); u++) {
      if (u!=j && u!=k) {
	cout << "Working on face #" << u << endl;
	gen X0i = evali(faces[u].X0);
	gen X1i = evali(faces[u].X1);
	
	//	cout << "X0i = " << X0i << endl;
	//	cout << "X1i = " << X1i << endl;

	gen a = conj(X0i,&ct)*Hn*W0n;
	gen b = conj(X0i,&ct)*Hn*V0n;
	gen c = conj(X1i,&ct)*Hn*W0n;
	gen d = conj(X1i,&ct)*Hn*V0n;
	gen den = a*conj(a,&ct)-c*conj(c,&ct);
	
	/*	cout << "a=" << a << endl;
	cout << "b=" << b << endl;
	cout << "c=" << c << endl;
	cout << "d=" << d << endl;*/
	den = re(den,&ct);

	if (is_greater(zeronumber,_left(den,&ct),&ct) && is_greater(_right(den,&ct),zeronumber,&ct)) {
	  cerr << "Interval for denominator contains 0, maybe a line?" << endl;	  
	}
	else {
	  gen ctr = (conj(c,&ct)*d-conj(a,&ct)*b)/den;
	  cout << "ctr=" << ctr << endl;
	  gen radsq = re( ctr*conj(ctr,&ct) , &ct ) - re( b*conj(b,&ct)-d*conj(d,&ct) , &ct )/den;
	  if (is_greater(_left(radsq,&ct),zeronumber,&ct)) {
	    gen rad = sqrt(radsq,&ct);
	    cout << "rad = " << rad << endl;
	    if (is_greater(1000,rad,&ct)) {
	      file << "draw(circle(("<<_left(re(ctr,&ct),&ct)<<","<<_left(im(ctr,&ct),&ct)<<"),"<<_left(rad,&ct)<<"),black);" << endl;
	      gen tee = ctr*(1-rad/abs(ctr,&ct));
	      /*	      if (u==1) {
		file << "label(\"$1,14$\",(" << re(tee,&ct) << "," << im(tee,&ct) << "),right);" << endl;
	      }
	      if (u==2) {
		file << "label(\"$2,7$\",(" << re(tee,&ct) << "," << im(tee,&ct)+0.1 << "));" << endl;
		}*/
	      file << "label(\"$" << u << "$\",(" << re(tee,&ct) << "," << im(tee,&ct) << "),black+fontsize(8));" << endl;	      
	    }
	  }
	  else {
	    if (is_greater(zeronumber,_right(radsq,&ct),&ct)) {
	      cout << "No intersection, radius of circle is negative" << endl;
	    }
	    else {
	      cout << "Doubtful intersection, the interval for radius contains 0" << endl;
	      cout << "Will simply draw nothing..." << endl;
	    }
	  }
	}
      } 
    }
    // then draw the geodesic polygon through vertices...
    cout << "Drawing polygon" << endl;
    vector<gen> verts_num;
    gen ii = gen("exp(pi*i/2)",&ct);
    gen fa = gen("180/pi",&ct);
    for (int u=0; u<ridges[i].verts.size(); u++) {
      gen X = evali(vertices[ridges[i].verts[u]].coords);
      gen rat = -(conj(X,&ct)*Hn*W0n)/(conj(X,&ct)*Hn*V0n);
      verts_num.push_back(_left(re(rat,&ct),&ct)+ii*_left(im(rat,&ct),&ct));
    }
    int si = verts_num.size();
    /*    for (int u = 0; u<si; u++) {
      file << "label(\"$" << u << "$\",(" << re(verts_num[u],&ct) << "," << im(verts_num[u],&ct) << "));" << endl;
      }*/

    file << "path pol = ";
    for (int u=0; u<si; u++) {
      gen alp = verts_num[u];
      gen bet = verts_num[(u+1)%si];

      gen den = conj(alp,&ct)*bet - alp*conj(bet,&ct);
      gen ctr = (alp*(alp*conj(alp,&ct)-bet*conj(bet,&ct)) + (bet-alp)*(1+alp*conj(alp,&ct)))/den;

      gen rad = sqrt(ctr*conj(ctr,&ct)-1,&ct);
      //      cout << "ctr: " << ctr << ",   ns = " << re(ctr*conj(ctr,&ct),&ct) << endl;
      //      cout << "rad: " << rad  << endl;
      gen u1 = alp-ctr;
      gen u3 = bet-ctr;
      gen th1 = arg(u1,&ct)*fa;
      gen th2 = arg(u3,&ct)*fa;
      //     cout << "th1=" << th1 << endl;
      //     cout << "th2=" << th2 << endl;
      if (is_greater(th2,th1,&ct)) {
	if (is_greater(th2-th1,180,&ct)) {
	  th2 = th2-360;
	}
      }
      else {
	if (is_greater(th1-th2,180,&ct)) {
	  th1 = th1-360;
	}
      }
      file << "arc(("<<re(ctr,&ct)<<","<<im(ctr,&ct)<<"),"<<rad<<","<<th1<<","<<th2<<")--" << endl;      
    } 
    file << "cycle;" << endl;
    file << "filldraw(pol,gray);" << endl;
    file << "limits((-1.1,-1.1),(1.1,1.1),Crop);";
    file.close();
  }
  
}

bool PolytopeSmall::checkt1val(gen rur, gen par, gen t1v, int &res) {
  gen par_refined;
  if (checkt1valAttempt(rur,par,t1v,res)) {
    return true;
  }
  else {
    if (checkEqualt1val(rur,par,t1v)) {
      res = 0;
      return true;
    }
    else {
      unsigned savend = nd;
      bool done = false;
      while (nd<ndmax && !done) {
	// compute par_refined?
	//	cout << "rur[2]=" << rur[2] << endl;
	gen sols =_realroot(makesequence(rur[2],prec),&ct);
	int count = 0;
	int ksave;
	for (unsigned k=0; k<(*sols._VECTptr).size(); k++) {
	  //	  cout << "sols[k][0]=" << sols[k][0] << endl;
	  gen Jt = _intersect(makesequence(sols[k][0],par),&ct);
	  //	  cout << "J=" << J << endl;
	  if (Jt.type==3) {
	    count++;
	    ksave = k;
	  }
	}
	if (count!=1) {
	  cout << "Could not identify par from rur (count=" << count << ")" << endl;
	  return false;
	}
	else {
	  par_refined = sols[ksave][0];
	}
	done = checkt1valAttempt(rur, par_refined, t1v, res);
	if (!done) {
	  increasePrecision();
	}
      }
      nd = savend;
      lowerPrecision();
      return done;
    }
  }
}

bool PolytopeSmall::checkt1valAttempt(gen rur, gen par, gen t1v, int &res) {
  gen t1vi = evali(t1v);
  gen t1_from_rur = _horner(makesequence(rur[5],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);
  if (is_greater(_left(t1vi,&ct),_right(t1_from_rur,&ct),&ct)) {
    res = -1;
    return true;
  }
  else {
    if (is_greater(_left(t1_from_rur,&ct),_right(t1vi,&ct),&ct)) {
      res = 1;
      return true;
    }
    else {
      return false;
    }
  }
}

bool PolytopeSmall::checkEqualt1val(gen rur, gen par, gen t1v) {
  gen te = _subst(makesequence(t1v,s__IDNT_e,rur[4]/rur[3]),&ct);
  te = _simplify(te - rur[5]/rur[3],&ct);
  te = _numer(te,&ct);
  te = _rem(makesequence(te,rur[2],s__IDNT_e),&ct);
  if (operator_equal(te,zeronumber,&ct)) {
    return true;
  }
  else {
    // need to check if par is a root of te!
    gen g = _gcd(makesequence(te,rur[2]),&ct);
    gen sols = _realroot(makesequence(g,prec),&ct);
    int count = 0;
    bool fd = false;
    int k = 0;
    while (!fd && k<(*sols._VECTptr).size()) {
      gen rt;
      if (sols[k][0].type==8) {
	rt = convert_interval(sols[k][0],nd,&ct);      
      }
      else {
	rt = sols[k][0];
      }
      gen Jt = _intersect(makesequence(rt,par),&ct);
      fd = Jt.type==3;
      k++;
    }
    return fd;
  }
}

bool PolytopeSmall::checkt2val(gen rur, gen par, gen t2v, int &res) {
  gen par_refined;
  if (checkt2valAttempt(rur,par,t2v,res)) {
    return true;
  }
  else {
    if (checkEqualt2val(rur,par,t2v)) {
      res = 0;
      return true;
    }
    else {
      unsigned savend = nd;
      bool done = false;
      while (nd<ndmax && !done) {
	// compute par_refined?
	gen sols =_realroot(makesequence(rur[2],prec),&ct);
	int count = 0;
	int ksave;
	for (unsigned k=0; k<(*sols._VECTptr).size(); k++) {
	  gen rt;
	  if (sols[k][0].type==8) {
	    rt = convert_interval(sols[k][0],nd,&ct);      
	  }
	  else {
	    rt = sols[k][0];
	  }
	  gen Jt = _intersect(makesequence(rt,par),&ct);
	  if (Jt.type==3) {
	    count++;
	    ksave = k;
	  }
	}
	if (count!=1) {
	  cout << "Could not identify par from rur (count=" << count << ")" << endl;
	  return false;
	}
	else {
	  par_refined = sols[ksave][0];
	}
	done = checkt2valAttempt(rur, par_refined, t2v, res);
	if (!done) {
	  increasePrecision();
	}
      }
      nd = savend;
      lowerPrecision();
      return done;
    }
  }
}

bool PolytopeSmall::checkt2valAttempt(gen rur, gen par, gen t2v, int &res) {
  gen t2vi = evali(t2v);
  gen t2_from_rur = _horner(makesequence(rur[6],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);
  if (is_greater(_left(t2vi,&ct),_right(t2_from_rur,&ct),&ct)) {
    res = -1;
    return true;
  }
  else {
    if (is_greater(_left(t2_from_rur,&ct),_right(t2vi,&ct),&ct)) {
      res = 1;
      return true;
    }
    else {
      return false;
    }
  }
}

bool PolytopeSmall::checkEqualt2val(gen rur, gen par, gen t2v) {
  gen te = _subst(makesequence(t2v,s__IDNT_e,rur[4]/rur[3]),&ct);
  te = _simplify(te - rur[6]/rur[3],&ct);
  te = _numer(te,&ct);
  te = _rem(makesequence(te,rur[2],s__IDNT_e),&ct);
  if (operator_equal(te,zeronumber,&ct)) {
    return true;
  }
  else {
    // need to check if par is a root of te!
    gen g = _gcd(makesequence(te,rur[2]),&ct);
    //    cout << "g=" << g << endl;
    gen sols = _realroot(makesequence(g,prec),&ct);
    int count = 0;
    bool fd = false;
    int k = 0;
    while (!fd && k<(*sols._VECTptr).size()) {
      gen rt;
      if (sols[k][0].type==8) {
	rt = convert_interval(sols[k][0],nd,&ct);      
      }
      else {
	rt = sols[k][0];
      }
      gen Jt = _intersect(makesequence(rt,par),&ct);
      fd = Jt.type==3;
      k++;
    }
    return fd;
  }
}

bool PolytopeSmall::checkt3val(gen rur, gen par, gen t3v, gen vl, gen wl, gen vt, int &res) {
  gen par_refined;
  if (checkt3valAttempt(rur,par,t3v,vl,wl,vt,res)) {
    return true;
  }
  else {
    if (checkEqualt3val(rur,par,t3v,vl,wl,vt)) {
      res = 0;
      return true;
    }
    else {
      unsigned savend = nd;
      bool done = false;
      while (nd<ndmax && !done) {
	gen sols =_realroot(makesequence(rur[2],prec),&ct);
	int count = 0;
	int ksave;
	for (unsigned k=0; k<(*sols._VECTptr).size(); k++) {
	  gen rt;
	  if (sols[k][0].type==8) {
	    rt = convert_interval(sols[k][0],nd,&ct);      
	  }
	  else {
	    rt = sols[k][0];
	  }
	  gen Jt = _intersect(makesequence(rt,par),&ct);
	  if (Jt.type==3) {
	    count++;
	    ksave = k;
	  }
	}
	if (count!=1) {
	  cout << "Could not identify par from rur (count=" << count << ")" << endl;
	  return false;
	}
	else {
	  par_refined = sols[ksave][0];
	}
	done = checkt3valAttempt(rur, par_refined, t3v,vl,wl,vt,res);
	if (!done) {
	  increasePrecision();
	}
      }
      nd = savend;
      lowerPrecision();
      return done;
    }
  }
}

bool PolytopeSmall::checkt3valAttempt(gen rur, gen par, gen t3v, gen vl, gen wl, gen vt, int &res) {

  gen svvi = _horner(makesequence(rur[5],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);
  gen t1vi = _horner(makesequence(rur[5],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);
  gen t2vi = _horner(makesequence(rur[6],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);

  gen alp = -Inn(vt,vl);
  gen bet = Inn(vt,wl);
  alp = (alp*myconj(bet)+myconj(alp)*bet)/2;
  bet = bet*myconj(bet);
  red(alp);
  red(bet);
  alp = convertToReal(alp);
  bet = convertToReal(bet);

  alp = _subst(makesequence(alp,s__IDNT_e,svvi),&ct);
  alp = _subst(makesequence(alp,t1,t1vi),&ct);
  alp = _subst(makesequence(alp,t2,t2vi),&ct);
  bet = _subst(makesequence(bet,s__IDNT_e,svvi),&ct);
  bet = _subst(makesequence(bet,t1,t1vi),&ct);
  bet = _subst(makesequence(bet,t2,t2vi),&ct);
  
  gen t3_from_rur = alp/bet;

  gen t3vi = evali(t3v);

  if (is_greater(_left(t3vi,&ct),_right(t3_from_rur,&ct),&ct)) {
    res = -1;
    return true;
  }
  else {
    if (is_greater(_left(t3_from_rur,&ct),_right(t3vi,&ct),&ct)) {
      res = 1;
      return true;
    }
    else {
      return false;
    }
  }
}

bool PolytopeSmall::checkEqualt3val(gen rur, gen par, gen t3v, gen vl, gen wl, gen vt) {

  gen alp = -Inn(vt,vl);
  gen bet = Inn(vt,wl);
  alp = (alp*myconj(bet)+myconj(alp)*bet)/2;
  bet = bet*myconj(bet);
  red(alp);
  red(bet);
  alp = convertToReal(alp);
  bet = convertToReal(bet);

  alp = _subst(makesequence(alp,s__IDNT_e,rur[4]/rur[3]),&ct);
  alp = _subst(makesequence(alp,t1,rur[5]/rur[3]),&ct);
  alp = _subst(makesequence(alp,t2,rur[6]/rur[3]),&ct);

  bet = _subst(makesequence(bet,s__IDNT_e,rur[4]/rur[3]),&ct);
  bet = _subst(makesequence(bet,t1,rur[5]/rur[3]),&ct);
  bet = _subst(makesequence(bet,t2,rur[6]/rur[3]),&ct);

  gen te = alp/bet - _subst(makesequence(t3v,s__IDNT_e,rur[4]/rur[3]),&ct);
  te = _numer(te,&ct);
  te = _rem(makesequence(te,rur[2],s__IDNT_e),&ct);

  if (operator_equal(te,zeronumber,&ct)) {
    return true;
  }
  else {
    // need to check if par is a root of te!
    gen g = _gcd(makesequence(te,rur[2]),&ct);
    gen sols = _realroot(makesequence(g,prec),&ct);
    int count = 0;
    bool fd = false;
    int k = 0;
    while (!fd && k<(*sols._VECTptr).size()) {
      gen rt;
      if (sols[k][0].type==8) {
	rt = convert_interval(sols[k][0],nd,&ct);      
      }
      else {
	rt = sols[k][0];
      }
      gen Jt = _intersect(makesequence(rt,par),&ct);
      fd = Jt.type==3;
      k++;
    }
    return fd;
  }
}


bool PolytopeSmall::checkEqSign(gen rur, gen par, gen eq, int &res) {
  gen par_refined;
  if (checkEqSignAttempt(rur,par,eq,res)) {
    return true;
  }
  else {
    if (checkEqZero(rur,par,eq)) {
      res = 0;
      return true;
    }
    else {
      unsigned savend = nd;
      bool done = false;
      while (nd<ndmax && !done) {
	gen sols =_realroot(makesequence(rur[2],prec),&ct);
	int count = 0;
	int ksave;
	for (unsigned k=0; k<(*sols._VECTptr).size(); k++) {
	  gen rt;
	  if (sols[k][0].type==8) {
	    rt = convert_interval(sols[k][0],nd,&ct);      
	  }
	  else {
	    rt = sols[k][0];
	  }
	  gen Jt = _intersect(makesequence(rt,par),&ct);
	  if (Jt.type==3) {
	    count++;
	    ksave = k;
	  }
	}
	if (count!=1) {
	  cout << "Could not identify par from rur (count=" << count << ")" << endl;
	  return false;
	}
	else {
	  par_refined = sols[ksave][0];
	}
	done = checkEqSignAttempt(rur,par_refined,eq,res);
	if (!done) {
	  increasePrecision();
	}
      }
      nd = savend;
      lowerPrecision();
      return done;
    }
  }
}

bool PolytopeSmall::checkEqSignAttempt(gen rur, gen par, gen eq, int &res) {
  gen svvi = _horner(makesequence(rur[5],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);
  gen t1vi = _horner(makesequence(rur[5],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);
  gen t2vi = _horner(makesequence(rur[6],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);

  gen eq_eval = _subst(makesequence(eq,s__IDNT_e,svvi),&ct);
  eq_eval = _subst(makesequence(eq_eval,t1,t1vi),&ct);
  eq_eval = _subst(makesequence(eq_eval,t2,t2vi),&ct);

  if (is_greater(zeronumber,_right(eq_eval,&ct),&ct)) {
    res = -1;
    return true;
  }
  else {
    if (is_greater(_left(eq_eval,&ct),zeronumber,&ct)) {
      res = 1;
      return true;
    }
    else {
      return false;
    }
  }
}

bool PolytopeSmall::checkEqZero(gen rur, gen par, gen eq) {
  gen te = _subst(makesequence(eq,s__IDNT_e,rur[4]/rur[3]),&ct);
  te = _subst(makesequence(te,t1,rur[5]/rur[3]),&ct);
  te = _subst(makesequence(te,t2,rur[6]/rur[3]),&ct);
  te = _numer(te,&ct);
  te = _rem(makesequence(te,rur[2],s__IDNT_e),&ct);

  if (operator_equal(te,zeronumber,&ct)) {
    return true;
  }
  else {
    // need to check if par is a root of te!
    gen g = _gcd(makesequence(te,rur[2]),&ct);
    gen sols = _realroot(makesequence(g,prec),&ct);
    int count = 0;
    bool fd = false;
    int k = 0;
    while (!fd && k<(*sols._VECTptr).size()) {
      gen rt;
      if (sols[k][0].type==8) {
	rt = convert_interval(sols[k][0],nd,&ct);      
      }
      else {
	rt = sols[k][0];
      }
      gen Jt = _intersect(makesequence(rt,par),&ct);
      fd = Jt.type==3;
      k++;
    }
    return fd;
  }
}

void PolytopeSmall::importFaceData() {
  cout << "Import face data (I will need to compute the sides of the prisms, this may take some time)..." << endl;
  /*  string filename = "faceData_4s5.chg";
  ifstream file(filename.c_str());*/
  ifstream file("faceData/faceData_4s5.chg");
  if (file) {
    string line;
    int count = 0;
    while (getline(file,line)) {
      int n = (int) line.at(1) - 48;
      //      cout << "#" << count << ": ";
      //      cout << n << "-gon" << endl;
      count++;
      getline(file,line);
      vector<int> m0;
      string s = "";
      for (unsigned l=0; l<line.length(); l++) {
	if (line.at(l)==',') {
	  int ne = atoi(s.c_str());
	  m0.push_back(ne);
	  s = "";
	}
	else {
	  s = s + line.at(l);
	}
      }
      int ne = atoi(s.c_str());
      m0.push_back(ne);

      //      cout << "Mirror: " << convertword(m0) << endl;
      vector<vector<int> > s0;
      for (unsigned k=0; k!=n; k++) {
	getline(file,line);
	vector<int> side;
	string s = "";
	for (unsigned l=0; l<line.length(); l++) {
	  if (line.at(l)==',') {
	    int ne = atoi(s.c_str());
	    side.push_back(ne);
	    s = "";
	  }
	  else {
	    s = s + line.at(l);
	  }
	}
	int ne = atoi(s.c_str());
	side.push_back(ne);
	s0.push_back(side);
      }
      for (unsigned u=0; u<s0.size(); u++) {
	//	cout << "Side #" << u << ": " << convertword(s0[u]) << endl;
      }
      Prism A;
      gen X = identifyMirror(m0);      
      vector<int> w;
      /*  int mi_n = m0.size()/2;
	  for (int k=0; k<mi_n; k++) {
	if (m0[k]!=-m0[m0.size()-1-k]) {
	  cerr << "Not anti-palyndromic (mirror)!" << endl;
	  exit(1);
	}
	w.push_back(m0[k]);
      }
      gen X;
      X = _col(makesequence(Id,abs(m0[mi_n])-1),&ct);*/
      evalWord(w,X);
      A.mirror = WVector(X,m0);
      //      cout << "Done with mirror" << endl;
      for (unsigned u=0; u<s0.size(); u++) {
	/*	cout << "s0[" << u << "]: " << s0[u] << endl;
	int s_n = (s0[u].size()-1)/2;
	//	cout << "s_n=" << s_n << endl;
	vector<int> wu;
	for (int k=0; k<s_n; k++) {
	  if (s0[u][k]!=-s0[u][s0[u].size()-1-k]) {
	    cerr << "Not anti-palyndromic (side #" << k << ")!" << endl;
	    cout << "k=" << k << ": " << s0[u][k] << " VS " << (-s0[u][s0[u].size()-1-k]) << endl;
	    exit(1);
	  }
	  else {
	    cout << "k=" << k << ": " << s0[u][k] << " pal ok" << endl;
	  }
	  wu.push_back(s0[u][k]);
	}
	//	cout << "wu=" << convertword(wu) << endl;
	gen X;
	X = _col(makesequence(Id,abs(s0[u][s_n])-1),&ct);
	evalWord(wu,X);*/
	//	cout << "Done with eval word, X=" << X << endl;
	gen X = identifyMirror(s0[u]);
	WVector si = WVector(X,s0[u]);
	A.sides.push_back(si);
      }
      vector<int> ww;
      gen apc = BoxProd(A.sides[0].coords,A.sides[1].coords); 
      WVector ap(apc,ww);
      A.apex = ap;
      A.apexProjection = Proj(apc,A.mirror.coords);
      A.cspine = BoxProd(A.apex.coords,A.apexProjection);
      //      cout << " Done with prism" << endl;
      faces.push_back(A);
    }
    cout << endl;
    cout << "Done importing data for " << faces.size() << " faces!" << endl;
    for (unsigned k=0; k<faces.size(); k++) {
      cout << "#" << k << ": " << faces[k] << endl;
    }
  }
  else {
    cout << "Problem opening file" << endl;
  }
}

gen PolytopeSmall::identifyMirror(vector<int> w) {
  gen M = evalWord(w) - mult*Id;
  matred(M);
  gen d = det(M);
  red(d);
  if (!operator_equal(d,zeronumber,&ct)) {
    cerr << "Problem finding mirror from matrix!" << endl;
    exit(1);
  }
  gen te1 = _row(makesequence(M,0),&ct);
  gen te2 = _row(makesequence(M,1),&ct);
  vecteur tec;
  for (unsigned k=0; k<3; k++) {
    gen te = te1[(k+1)%3]*te2[(k+2)%3]-te1[(k+2)%3]*te2[(k+1)%3];
    red(te);
    tec.push_back(te);
  }
  gen tee = gen(tec);
  return tee;
}

bool PolytopeSmall::checkSide() {
  vector<int> inds;
  for (int k=0; k<faces.size(); k++) {
    inds.push_back(k);
  }
  while (inds.size()>0) {
    //    cout << "inds: " << inds << endl;
    int k = inds[0];
    bool foundbuddy = false;
    int u = 0;
    while (!foundbuddy && u<inds.size()-1) {
      u++;
      gen X = BoxProd(faces[k].cspine,faces[inds[u]].cspine);
      if (!operator_equal(X,zerovector,&ct)) {
	int sx;
	if (checkSign(SqNorm(X),sx)) {
	  foundbuddy = sx<0;
	  if (foundbuddy) {
	    //	    cout << "Face " << k << " is buddy with face " << inds[u] << endl;
	    faces[k].X0 = X;
	    faces[k].X1 = reflectAcrossSpine(X,faces[k]);
	    faces[inds[u]].X0 = X;
	    faces[inds[u]].X1 = reflectAcrossSpine(X,faces[inds[u]]); 
	  }
	}
      }
    }
    if (!foundbuddy) {
      // may have to use other faces as buddies...
      u=-1;
      while (!foundbuddy && u!=faces.size()-1) {
	u++;
	gen X = BoxProd(faces[k].cspine,faces[u].cspine);
	if (!operator_equal(X,zerovector,&ct)) {
	  int sx;
	  if (checkSign(SqNorm(X),sx)) {
	    foundbuddy = sx<0;
	    if (foundbuddy) {
	      //	      cout << "Face " << k << " is buddy with face " << u << endl;
	      faces[k].X0 = X;
	      faces[k].X1 = reflectAcrossSpine(X,faces[k]);
	    }
	  }
	}
      }
      if (!foundbuddy) {
	cerr << "Face #" << k << " is not coequidistant from anyone..." << endl;
	cerr << " this may not be a serious issue, but I will stop now." << endl;
	exit(1);
      }
      else {
	inds.erase(inds.begin());
      }
    }
    else {
      inds.erase(inds.begin()+u);
      if (u!=0) {
	inds.erase(inds.begin());
      }
      else {
	cerr << "Trouble in checkSide()" << endl;
	exit(1);
      }
    }
  }
  gen Hn = evali(H);
  gen p0n = evalf(p0,20,&ct);
  bool res = true;
  for (unsigned k=0; k<faces.size() && res; k++) {
    gen u1 = evali(faces[k].X0);
    gen v1 = evali(faces[k].X1);
    gen te = abs( conj(p0n,&ct)*Hn*u1 ) - abs( conj(p0n,&ct)*Hn*v1 );
    res = is_greater(zeronumber,_right(te,&ct),&ct); 
    if (!res) {
      cout << "Test face " << k << ": " << te << endl;
    }
  }
  return res;

}
